# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

# TODO(tehrengruber): This file contains to many different components. Split
#  into components for each dialect.


from __future__ import annotations

import dataclasses
import functools
import time
import types
import typing
import warnings
from collections.abc import Callable, Sequence
from typing import Any, Generic, Optional, TypeVar

from gt4py import eve
from gt4py._core import definitions as core_defs
from gt4py.eve import extended_typing as xtyping, utils as eve_utils
from gt4py.eve.extended_typing import Self, override
from gt4py.next import (
    allocators as next_allocators,
    backend as next_backend,
    common,
    config,
    embedded as next_embedded,
    errors,
    metrics,
    utils,
)
from gt4py.next.embedded import operators as embedded_operators
from gt4py.next.ffront import (
    field_operator_ast as foast,
    foast_to_gtir,
    past_process_args,
    signature,
    stages as ffront_stages,
    transform_utils,
    type_specifications as ts_ffront,
)
from gt4py.next.ffront.gtcallable import GTCallable
from gt4py.next.iterator import ir as itir
from gt4py.next.otf import arguments, compiled_program, stages, toolchain
from gt4py.next.type_system import type_info, type_specifications as ts, type_translation


DEFAULT_BACKEND: next_backend.Backend | None = None


def _implicit_offset_providers_from_types(types: list[ts.TypeSpec]) -> common.OffsetProvider:
    """
    Add all implicit offset providers.
    Each dimension implicitly defines an offset provider such that we can allow syntax like::
        field(TDim + 1)
    This function adds these implicit offset providers.
    """
    # TODO(tehrengruber): We add all dimensions here regardless of whether they are cartesian
    #  or unstructured. While it is conceptually fine, but somewhat meaningless,
    #  to do something `Cell+1` the GTFN backend for example doesn't support these. We should
    #  find a way to avoid adding these dimensions, but since we don't have the grid type here
    #  and since the dimensions don't this information either, we just add all dimensions here
    #  and filter them out in the backends that don't support this.
    implicit_offset_provider = {}
    for type_ in types:
        if isinstance(type_, ts.FieldType):
            for dim in type_.dims:
                if dim.kind in (common.DimensionKind.HORIZONTAL, common.DimensionKind.VERTICAL):
                    implicit_offset_provider.update(
                        {common.dimension_to_implicit_offset(dim.value): dim}
                    )
    return implicit_offset_provider


# TODO(tehrengruber): Decide if and how programs can call other programs. As a
#  result Program could become a GTCallable.
@dataclasses.dataclass(frozen=True)
class Program:
    """
    Construct a program object from a PAST node.

    A call to the resulting object executes the program as expressed
    by the PAST node.

    Attributes:
        past_node: The node representing the program.
        closure_vars: Mapping of externally defined symbols to their respective values.
            For example, referenced global and nonlocal variables.
        backend: The backend to be used for code generation.
        definition: The Python function object corresponding to the PAST node.
        grid_type: The grid type (cartesian or unstructured) to be used. If not explicitly given
            it will be deduced from actually occurring dimensions.
        connectivities: A dictionary holding static/compile-time information about the offset providers.
            For now, it is used for ahead of time compilation in DaCe orchestrated programs,
            i.e. DaCe programs that call GT4Py Programs -SDFGConvertible interface-.
    """

    definition_stage: ffront_stages.ProgramDefinition
    backend: Optional[next_backend.Backend]
    connectivities: Optional[
        common.OffsetProvider
    ]  # TODO(ricoh): replace with common.OffsetProviderType once the temporary pass doesn't require the runtime information
    enable_jit: bool | None
    static_params: (
        Sequence[str] | None
    )  # if the user requests static params, they will be used later to initialize CompiledPrograms

    _extended_offset_provider_cache: eve_utils.CustomMapping = dataclasses.field(
        default_factory=lambda: eve_utils.CustomMapping(common.hash_offset_provider_unsafe),
        kw_only=True,
        init=False,
    )  # cache the extended dictionary of offset providers that includes the implicitly defined ones

    @classmethod
    def from_function(
        cls,
        definition: types.FunctionType,
        backend: next_backend.Backend | None,
        grid_type: common.GridType | None = None,
        enable_jit: bool | None = None,
        static_params: Sequence[str] | None = None,
        connectivities: Optional[
            common.OffsetProvider
        ] = None,  # TODO(ricoh): replace with common.OffsetProviderType once the temporary pass doesn't require the runtime information
    ) -> Program:
        program_def = ffront_stages.ProgramDefinition(definition=definition, grid_type=grid_type)
        return cls(
            definition_stage=program_def,
            backend=backend,
            connectivities=connectivities,
            enable_jit=enable_jit,
            static_params=static_params,
        )

    # needed in testing
    @property
    def definition(self) -> types.FunctionType:
        return self.definition_stage.definition

    @functools.cached_property
    def past_stage(self) -> ffront_stages.PRG:
        # backwards compatibility for backends that do not support the full toolchain
        no_args_def = toolchain.CompilableProgram(
            self.definition_stage, arguments.CompileTimeArgs.empty()
        )
        return self._frontend_transforms.func_to_past(no_args_def).data

    # TODO(ricoh): linting should become optional, up to the backend.
    def __post_init__(self) -> None:
        no_args_past = toolchain.CompilableProgram(
            self.past_stage, arguments.CompileTimeArgs.empty()
        )
        _ = self._frontend_transforms.past_lint(no_args_past).data

    @property
    def __name__(self) -> str:
        return self.definition_stage.definition.__name__

    @functools.cached_property
    def __gt_allocator__(
        self,
    ) -> next_allocators.FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]:
        if self.backend:
            return self.backend.__gt_allocator__
        else:
            raise RuntimeError(f"Program '{self}' does not have a backend set.")

    @property
    def _frontend_transforms(self) -> next_backend.Transforms:
        if self.backend is None:
            return next_backend.DEFAULT_TRANSFORMS
        # TODO(tehrengruber): This class relies heavily on `self.backend.transforms` being
        #  a `next_backend.Transforms`, but the backend type annotation does not reflect that.
        assert isinstance(self.backend.transforms, next_backend.Transforms)
        return self.backend.transforms

    def with_backend(self, backend: next_backend.Backend) -> Program:
        return dataclasses.replace(self, backend=backend)

    def with_connectivities(
        self,
        connectivities: common.OffsetProvider,  # TODO(ricoh): replace with common.OffsetProviderType once the temporary pass doesn't require the runtime informatio
    ) -> Program:
        return dataclasses.replace(self, connectivities=connectivities)

    def with_grid_type(self, grid_type: common.GridType) -> Program:
        return dataclasses.replace(
            self, definition_stage=dataclasses.replace(self.definition_stage, grid_type=grid_type)
        )

    def with_static_params(self, *static_params: str | None) -> Program:
        if not static_params or (static_params == (None,)):
            _static_params = None
        else:
            assert all(p is not None for p in static_params)
            _static_params = typing.cast(tuple[str], static_params)
        return dataclasses.replace(
            self,
            static_params=_static_params,
        )

    def with_bound_args(self, **kwargs: Any) -> ProgramWithBoundArgs:
        """
        Bind scalar, i.e. non field, program arguments.

        Example (pseudo-code):

        >>> import gt4py.next as gtx
        >>> @gtx.program  # doctest: +SKIP
        ... def program(condition: bool, out: gtx.Field[[IDim], float]):  # noqa: F821 [undefined-name]
        ...     sample_field_operator(condition, out=out)  # noqa: F821 [undefined-name]

        Create a new program from `program` with the `condition` parameter set to `True`:

        >>> program_with_bound_arg = program.with_bound_args(condition=True)  # doctest: +SKIP

        The resulting program is equivalent to

        >>> @gtx.program  # doctest: +SKIP
        ... def program(condition: bool, out: gtx.Field[[IDim], float]):  # noqa: F821 [undefined-name]
        ...     sample_field_operator(condition=True, out=out)  # noqa: F821 [undefined-name]

        and can be executed without passing `condition`.

        >>> program_with_bound_arg(out, offset_provider={})  # doctest: +SKIP
        """
        for key in kwargs.keys():
            if all(key != param.id for param in self.past_stage.past_node.params):
                raise TypeError(f"Keyword argument '{key}' is not a valid program parameter.")

        return ProgramWithBoundArgs(
            bound_args=kwargs,
            **{
                field.name: getattr(self, field.name)
                for field in dataclasses.fields(self)
                if field.init
            },
        )

    @functools.cached_property
    def _all_closure_vars(self) -> dict[str, Any]:
        return transform_utils._get_closure_vars_recursively(self.past_stage.closure_vars)

    @functools.cached_property
    def gtir(self) -> itir.Program:
        no_args_past = toolchain.CompilableProgram(
            data=ffront_stages.PastProgramDefinition(
                past_node=self.past_stage.past_node,
                closure_vars=self.past_stage.closure_vars,
                grid_type=self.definition_stage.grid_type,
            ),
            args=arguments.CompileTimeArgs.empty(),
        )
        return self._frontend_transforms.past_to_itir(no_args_past).data

    @functools.cached_property
    def _implicit_offset_provider(self) -> common.OffsetProvider:
        return _implicit_offset_providers_from_types(
            [param.type for param in self.past_stage.past_node.params]
        )

    @functools.cached_property
    def _compiled_programs(self) -> compiled_program.CompiledProgramsPool:
        if self.backend is None or self.backend == eve.NOTHING:
            raise RuntimeError("Cannot compile a program without backend.")

        if self.static_params is None:
            object.__setattr__(self, "static_params", ())

        program_type = self.past_stage.past_node.type
        assert isinstance(program_type, ts_ffront.ProgramType)
        return compiled_program.CompiledProgramsPool(
            backend=self.backend,
            definition_stage=self.definition_stage,
            program_type=program_type,
            static_params=self.static_params,
        )

    def _extend_offset_provider(
        self,
        offset_provider: common.OffsetProvider,
    ) -> common.OffsetProvider:
        try:
            extended_op = self._extended_offset_provider_cache[offset_provider]
        except KeyError:
            extended_op = {**self._implicit_offset_provider, **offset_provider}
            self._extended_offset_provider_cache[offset_provider] = extended_op
        return extended_op

    def __call__(
        self,
        *args: Any,
        offset_provider: common.OffsetProvider | None = None,
        enable_jit: bool | None = None,
        **kwargs: Any,
    ) -> None:
        if offset_provider is None:
            offset_provider = {}
        if enable_jit is None:
            enable_jit = (
                self.enable_jit if self.enable_jit is not None else config.ENABLE_JIT_DEFAULT
            )
        program_name = (
            f"{self.__name__}[{getattr(self.backend, 'name', '<embedded>')}]"
            if config.COLLECT_METRICS_LEVEL
            else ""
        )
        with metrics.metrics_collection(program_name) as metrics_collection:
            collect_metrics = metrics_collection is not None and (
                config.COLLECT_METRICS_LEVEL >= metrics.INFO
            )
            if collect_metrics:
                start = time.time()

            if __debug__:
                # TODO: remove or make dependency on self.past_stage optional
                past_process_args._validate_args(
                    self.past_stage.past_node,
                    arg_types=[type_translation.from_value(arg) for arg in args],
                    kwarg_types={k: type_translation.from_value(v) for k, v in kwargs.items()},
                )

            offset_provider = self._extend_offset_provider(offset_provider)
            if self.backend is not None:
                self._compiled_programs(
                    *args, **kwargs, offset_provider=offset_provider, enable_jit=enable_jit
                )
            else:
                # embedded
                warnings.warn(
                    UserWarning(
                        f"Field View Program '{self.definition_stage.definition.__name__}': Using Python execution, consider selecting a performance backend."
                    ),
                    stacklevel=2,
                )
                with next_embedded.context.update(offset_provider=offset_provider):
                    self.definition_stage.definition(*args, **kwargs)

            if metrics_collection is not None and collect_metrics:
                metrics_collection[metrics.TOTAL_METRIC].add_sample(time.time() - start)

    def compile(
        self,
        offset_provider: common.OffsetProviderType
        | common.OffsetProvider
        | list[common.OffsetProviderType | common.OffsetProvider]
        | None = None,
        enable_jit: bool | None = None,
        **static_args: list[xtyping.MaybeNestedInTuple[core_defs.Scalar]],
    ) -> Self:
        """
        Compiles the program for the given combination of static arguments and offset provider type.

        Note: Unlike `with_...` methods, this method does not return a new instance of the program,
        but adds the compiled variants to the current program instance.
        """
        # TODO(havogt): we should reconsider if we want to return a new program on `compile` (and
        # rename to `with_static_args` or similar) once we have a better understanding of the
        # use-cases.

        if enable_jit is not None:
            object.__setattr__(self, "enable_jit", enable_jit)
        if self.static_params is None:
            object.__setattr__(self, "static_params", tuple(static_args.keys()))
        if self.connectivities is None and offset_provider is None:
            raise ValueError(
                "Cannot compile a program without connectivities / OffsetProviderType."
            )
        if not all(isinstance(v, list) for v in static_args.values()):
            raise TypeError(
                "Please provide the static arguments as lists."
            )  # To avoid confusion with tuple args

        offset_provider = self.connectivities if offset_provider is None else offset_provider
        if not isinstance(offset_provider, list):
            offset_provider = [offset_provider]  # type: ignore[list-item] # cleanup offset_provider vs offset_provider_type

        assert all(
            common.is_offset_provider(op) or common.is_offset_provider_type(op)
            for op in offset_provider
        )
        offset_provider = [self._extend_offset_provider(op) for op in offset_provider]  # type: ignore[arg-type] # cleanup offset_provider vs offset_provider_type

        self._compiled_programs.compile(offset_providers=offset_provider, **static_args)
        return self

    def freeze(self) -> FrozenProgram:
        if self.backend is None:
            raise ValueError("Can not freeze a program without backend (embedded execution).")
        return FrozenProgram(
            self.definition_stage if self.definition_stage else self.past_stage,
            backend=self.backend,
        )


@dataclasses.dataclass(frozen=True)
class FrozenProgram:
    """
    Simplified program instance, which skips the whole toolchain after the first execution.

    Does not work in embedded execution.
    """

    program: ffront_stages.DSL_PRG | ffront_stages.PRG
    backend: next_backend.Backend
    _compiled_program: Optional[stages.CompiledProgram] = dataclasses.field(
        init=False, default=None
    )

    def __post_init__(self) -> None:
        if self.backend is None:
            raise ValueError("Can not JIT-compile programs without backend (embedded execution).")

    @property
    def definition(self) -> types.FunctionType:
        # `PastProgramDefinition` doesn't have `definition`
        assert isinstance(self.program, ffront_stages.ProgramDefinition)
        return self.program.definition

    def with_backend(self, backend: next_backend.Backend) -> FrozenProgram:
        return self.__class__(program=self.program, backend=backend)

    def with_grid_type(self, grid_type: common.GridType) -> FrozenProgram:
        return self.__class__(
            program=dataclasses.replace(self.program, grid_type=grid_type), backend=self.backend
        )

    def jit(
        self, *args: Any, offset_provider: common.OffsetProvider, **kwargs: Any
    ) -> stages.CompiledProgram:
        return self.backend.jit(self.program, *args, offset_provider=offset_provider, **kwargs)

    def __call__(self, *args: Any, offset_provider: common.OffsetProvider, **kwargs: Any) -> None:
        args, kwargs = signature.convert_to_positional(self.program, *args, **kwargs)

        if not self._compiled_program:
            super().__setattr__(
                "_compiled_program", self.jit(*args, offset_provider=offset_provider, **kwargs)
            )
        self._compiled_program(*args, offset_provider=offset_provider, **kwargs)  # type: ignore[misc] # _compiled_program is not None


try:
    from gt4py.next.program_processors.runners.dace.program import (  # type: ignore[assignment]
        Program,
    )
except ImportError:
    pass


# TODO(tehrengruber): This class does not follow the Liskov-Substitution principle as it doesn't
#  have a program definition. Revisit.
@dataclasses.dataclass(frozen=True)
class ProgramFromPast(Program):
    """
    This version of program has no DSL definition associated with it.

    PAST nodes can be built programmatically from field operators or from scratch.
    This wrapper provides the appropriate toolchain entry points.
    """

    past_stage: ffront_stages.PastProgramDefinition

    @override
    def __call__(
        self, *args: Any, offset_provider: Optional[common.OffsetProvider] = None, **kwargs: Any
    ) -> None:
        if self.backend is None:
            raise NotImplementedError(
                "Programs created from a PAST node (without a function definition) can not be executed in embedded mode"
            )

        if offset_provider is None:
            offset_provider = {}
        # TODO(ricoh): add test that does the equivalent of IDim + 1 in a ProgramFromPast
        self.backend(
            self.past_stage,
            *args,
            **(kwargs | {"offset_provider": {**offset_provider, **self._implicit_offset_provider}}),
        )

    # TODO(ricoh): linting should become optional, up to the backend.
    def __post_init__(self) -> None:
        self._frontend_transforms.past_lint(self.past_stage)  # type: ignore[arg-type] # ignored because the class has more TODO than code


@dataclasses.dataclass(frozen=True)
class ProgramWithBoundArgs(Program):
    bound_args: dict[str, float | int | bool] = dataclasses.field(default_factory=dict)

    @override
    def __call__(
        self, *args: Any, offset_provider: common.OffsetProvider | None = None, **kwargs: Any
    ) -> None:
        if offset_provider is None:
            offset_provider = {}
        type_ = self.past_stage.past_node.type
        assert isinstance(type_, ts_ffront.ProgramType)
        new_type = ts_ffront.ProgramType(
            definition=ts.FunctionType(
                kw_only_args={
                    k: v
                    for k, v in type_.definition.kw_only_args.items()
                    if k not in self.bound_args.keys()
                },
                pos_only_args=type_.definition.pos_only_args,
                pos_or_kw_args={
                    k: v
                    for k, v in type_.definition.pos_or_kw_args.items()
                    if k not in self.bound_args.keys()
                },
                returns=type_.definition.returns,
            )
        )

        arg_types = [type_translation.from_value(arg) for arg in args]
        kwarg_types = {k: type_translation.from_value(v) for k, v in kwargs.items()}

        try:
            # This error is also catched using `accepts_args`, but we do it manually here to give
            # a better error message.
            for name in self.bound_args.keys():
                if name in kwargs:
                    raise ValueError(f"Parameter '{name}' already set as a bound argument.")

            type_info.accepts_args(
                new_type, with_args=arg_types, with_kwargs=kwarg_types, raise_exception=True
            )
        except ValueError as err:
            bound_arg_names = ", ".join([f"'{bound_arg}'" for bound_arg in self.bound_args.keys()])
            raise TypeError(
                f"Invalid argument types in call to program '{self.past_stage.past_node.id}' with "
                f"bound arguments '{bound_arg_names}'."
            ) from err

        full_args = [*args]
        full_kwargs = {**kwargs}
        for index, param in enumerate(self.past_stage.past_node.params):
            if param.id in self.bound_args.keys():
                if index < len(full_args):
                    full_args.insert(index, self.bound_args[param.id])
                else:
                    full_kwargs[str(param.id)] = self.bound_args[param.id]

        return super().__call__(*tuple(full_args), offset_provider=offset_provider, **full_kwargs)

    @override
    def compile(
        self,
        offset_provider: common.OffsetProviderType
        | common.OffsetProvider
        | list[common.OffsetProviderType | common.OffsetProvider]
        | None = None,
        enable_jit: bool | None = None,
        **static_args: list[xtyping.MaybeNestedInTuple[core_defs.Scalar]],
    ) -> Self:
        raise NotImplementedError("Compilation of programs with bound arguments is not implemented")


@typing.overload
def program(definition: types.FunctionType) -> Program: ...


@typing.overload
def program(
    *,
    backend: next_backend.Backend | eve.NothingType | None,
    grid_type: common.GridType | None,
    enable_jit: bool | None,
    static_params: Sequence[str] | None,
    frozen: bool,
) -> Callable[[types.FunctionType], Program]: ...


def program(
    definition: types.FunctionType | None = None,
    *,
    # `NOTHING` -> default backend, `None` -> no backend (embedded execution)
    backend: next_backend.Backend | eve.NothingType | None = eve.NOTHING,
    grid_type: common.GridType | None = None,
    enable_jit: bool | None = None,  # only relevant if static_params are set
    static_params: Sequence[str] | None = None,
    frozen: bool = False,
) -> Program | FrozenProgram | Callable[[types.FunctionType], Program | FrozenProgram]:
    """
    Generate an implementation of a program from a Python function object.

    Examples:
        >>> @program  # noqa: F821 [undefined-name]  # doctest: +SKIP
        ... def program(in_field: Field[[TDim], float64], out_field: Field[[TDim], float64]):  # noqa: F821 [undefined-name]
        ...     field_op(in_field, out=out_field)
        >>> program(in_field, out=out_field)  # noqa: F821 [undefined-name]  # doctest: +SKIP

        >>> # the backend can optionally be passed if already decided
        >>> # not passing it will result in embedded execution by default
        >>> # the above is equivalent to
        >>> @program(backend="roundtrip")  # noqa: F821 [undefined-name]  # doctest: +SKIP
        ... def program(in_field: Field[[TDim], float64], out_field: Field[[TDim], float64]):  # noqa: F821 [undefined-name]
        ...     field_op(in_field, out=out_field)
        >>> program(in_field, out=out_field)  # noqa: F821 [undefined-name]  # doctest: +SKIP
    """

    def program_inner(definition: types.FunctionType) -> Program:
        program = Program.from_function(
            definition,
            backend=typing.cast(
                next_backend.Backend | None, DEFAULT_BACKEND if backend is eve.NOTHING else backend
            ),
            grid_type=grid_type,
            enable_jit=enable_jit,
            static_params=static_params,
        )
        if frozen:
            return program.freeze()  # type: ignore[return-value] # TODO(havogt): Should `FrozenProgram` be a `Program`?
        return program

    return program_inner if definition is None else program_inner(definition)


OperatorNodeT = TypeVar("OperatorNodeT", bound=foast.LocatedNode)


@dataclasses.dataclass(frozen=True)
class FieldOperator(GTCallable, Generic[OperatorNodeT]):
    """
    Construct a field operator object from a FOAST node.

    A call to the resulting object executes the field operator as expressed
    by the FOAST node and with the signature as if it would appear inside
    a program.

    Attributes:
        foast_node: The node representing the field operator.
        closure_vars: Mapping of names referenced in the field operator (i.e.
            globals, nonlocals) to their values.
        backend: The backend used for executing the field operator. Only used
            if the field operator is called directly, otherwise the backend
            specified for the program takes precedence.
        definition: The original Python function object the field operator
            was created from.
        grid_type: The grid type (cartesian or unstructured) to be used. If not explicitly given
            it will be deduced from actually occurring dimensions.
    """

    definition_stage: ffront_stages.FieldOperatorDefinition
    backend: Optional[next_backend.Backend]
    _program_cache: dict = dataclasses.field(
        init=False, default_factory=dict
    )  # init=False ensure the cache is not copied in calls to replace

    @classmethod
    def from_function(
        cls,
        definition: types.FunctionType,
        backend: Optional[next_backend.Backend],
        grid_type: Optional[common.GridType] = None,
        *,
        operator_node_cls: type[OperatorNodeT] = foast.FieldOperator,  # type: ignore[assignment] # TODO(ricoh): understand why mypy complains
        operator_attributes: Optional[dict[str, Any]] = None,
    ) -> FieldOperator[OperatorNodeT]:
        return cls(
            definition_stage=ffront_stages.FieldOperatorDefinition(
                definition=definition,
                grid_type=grid_type,
                node_class=operator_node_cls,
                attributes=operator_attributes or {},
            ),
            backend=backend,
        )

    # TODO(ricoh): linting should become optional, up to the backend.
    def __post_init__(self) -> None:
        """This ensures that DSL linting occurs at decoration time."""
        _ = self.foast_stage

    @functools.cached_property
    def foast_stage(self) -> ffront_stages.FoastOperatorDefinition:
        return self._frontend_transforms.func_to_foast(
            toolchain.CompilableProgram(
                data=self.definition_stage, args=arguments.CompileTimeArgs.empty()
            )
        ).data

    @property
    def __name__(self) -> str:
        return self.definition_stage.definition.__name__

    @property
    def definition(self) -> types.FunctionType:
        return self.definition_stage.definition

    @property
    def _frontend_transforms(self) -> next_backend.Transforms:
        if self.backend is None:
            return next_backend.DEFAULT_TRANSFORMS
        # TODO(tehrengruber): This class relies heavily on `self.backend.transforms` being
        #  a `next_backend.Transforms`, but the backend type annotation does not reflect that.
        assert isinstance(self.backend.transforms, next_backend.Transforms)
        return self.backend.transforms

    def __gt_type__(self) -> ts.CallableType:
        type_ = self.foast_stage.foast_node.type
        assert isinstance(type_, ts.CallableType)
        return type_

    def with_backend(self, backend: next_backend.Backend) -> FieldOperator:
        return dataclasses.replace(self, backend=backend)

    def with_grid_type(self, grid_type: common.GridType) -> FieldOperator:
        return dataclasses.replace(
            self, definition_stage=dataclasses.replace(self.definition_stage, grid_type=grid_type)
        )

    # TODO(tehrengruber): We can not use transforms from `self.backend` since this can be
    #  a different backend than the one of the program that calls this field operator. Just use
    #  the hard-coded lowering until this is cleaned up.
    def __gt_itir__(self) -> itir.FunctionDefinition:
        return foast_to_gtir.foast_to_gtir(self.foast_stage)

    # FIXME[#1582](tehrengruber): remove after refactoring to GTIR
    def __gt_gtir__(self) -> itir.FunctionDefinition:
        return foast_to_gtir.foast_to_gtir(self.foast_stage)

    def __gt_closure_vars__(self) -> dict[str, Any]:
        return self.foast_stage.closure_vars

    @functools.cached_property
    def _implicit_offset_provider(self) -> common.OffsetProvider:
        return _implicit_offset_providers_from_types(
            [param.type for param in self.foast_stage.foast_node.definition.params]
        )

    def as_program(self, compiletime_args: arguments.CompileTimeArgs) -> Program:
        foast_with_types = (
            toolchain.CompilableProgram(
                data=self.foast_stage,
                args=compiletime_args,
            ),
        )

        past_stage = self._frontend_transforms.field_view_op_to_prog.foast_to_past(  # type: ignore[attr-defined] # TODO(havogt): needs more work
            foast_with_types
        ).data
        return ProgramFromPast(
            definition_stage=None,  # type: ignore[arg-type] # ProgramFromPast needs to be fixed
            past_stage=past_stage,
            backend=self.backend,
            connectivities=None,
            enable_jit=False,  # TODO(havogt): revisit ProgramFromPast
            static_params=None,  # TODO(havogt): revisit ProgramFromPast
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if not next_embedded.context.within_valid_context() and self.backend is not None:
            # non embedded execution
            offset_provider = {
                **kwargs.pop("offset_provider", {}),
                **self._implicit_offset_provider,
            }
            if "out" not in kwargs:
                raise errors.MissingArgumentError(None, "out", True)
            out = kwargs.pop("out")
            if "domain" in kwargs:
                domain = common.domain(kwargs.pop("domain"))
                out = utils.tree_map(lambda f: f[domain])(out)

            args, kwargs = type_info.canonicalize_arguments(
                self.foast_stage.foast_node.type, args, kwargs
            )
            return self.backend(
                self.definition_stage,
                *args,
                out=out,
                offset_provider=offset_provider,
                **kwargs,
            )
        else:
            if not next_embedded.context.within_valid_context():
                # field_operator as program
                kwargs["offset_provider"] = {
                    **kwargs.pop("offset_provider", {}),
                    **self._implicit_offset_provider,
                }
            attributes = (
                self.definition_stage.attributes
                if self.definition_stage
                else self.foast_stage.attributes
            )
            if attributes is not None and any(
                has_scan_op_attribute := [
                    attribute in attributes for attribute in ["init", "axis", "forward"]
                ]
            ):
                assert all(has_scan_op_attribute)
                forward = attributes["forward"]
                init = attributes["init"]
                axis = attributes["axis"]
                op: embedded_operators.EmbeddedOperator = embedded_operators.ScanOperator(
                    self.definition_stage.definition, forward, init, axis
                )
            else:
                op = embedded_operators.EmbeddedOperator(self.definition_stage.definition)
            return embedded_operators.field_operator_call(op, args, kwargs)


@dataclasses.dataclass(frozen=True)
class FieldOperatorFromFoast(FieldOperator):
    """
    This version of the field operator does not have a DSL definition.

    FieldOperator AST nodes can be programmatically built, which may be
    particularly useful in testing and debugging.
    This class provides the appropriate toolchain entry points.
    """

    foast_stage: ffront_stages.FoastOperatorDefinition

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        assert self.backend is not None
        return self.backend(self.foast_stage, *args, **kwargs)


@typing.overload
def field_operator(
    definition: types.FunctionType,
    *,
    backend: next_backend.Backend | eve.NothingType | None,
    grid_type: common.GridType | None,
) -> FieldOperator[foast.FieldOperator]: ...


@typing.overload
def field_operator(
    *, backend: next_backend.Backend | eve.NothingType | None, grid_type: common.GridType | None
) -> Callable[[types.FunctionType], FieldOperator[foast.FieldOperator]]: ...


def field_operator(
    definition: types.FunctionType | None = None,
    *,
    backend: next_backend.Backend | eve.NothingType | None = eve.NOTHING,
    grid_type: common.GridType | None = None,
) -> (
    FieldOperator[foast.FieldOperator]
    | Callable[[types.FunctionType], FieldOperator[foast.FieldOperator]]
):
    """
    Generate an implementation of the field operator from a Python function object.

    Examples:
        >>> @field_operator  # doctest: +SKIP
        ... def field_op(in_field: Field[[TDim], float64]) -> Field[[TDim], float64]:  # noqa: F821 [undefined-name]
        ...     ...
        >>> field_op(in_field, out=out_field)  # noqa: F821 [undefined-name]  # doctest: +SKIP

        >>> # the backend can optionally be passed if already decided
        >>> # not passing it will result in embedded execution by default
        >>> @field_operator(backend="roundtrip")  # doctest: +SKIP
        ... def field_op(in_field: Field[[TDim], float64]) -> Field[[TDim], float64]:  # noqa: F821 [undefined-name]
        ...     ...
    """

    def field_operator_inner(definition: types.FunctionType) -> FieldOperator[foast.FieldOperator]:
        return FieldOperator.from_function(
            definition,
            typing.cast(
                next_backend.Backend | None, DEFAULT_BACKEND if backend is eve.NOTHING else backend
            ),
            grid_type,
        )

    return field_operator_inner if definition is None else field_operator_inner(definition)


@typing.overload
def scan_operator(
    definition: types.FunctionType,
    *,
    axis: common.Dimension,
    forward: bool,
    init: core_defs.Scalar,
    backend: next_backend.Backend | eve.NothingType | None,
    grid_type: common.GridType | None,
) -> FieldOperator[foast.ScanOperator]: ...


@typing.overload
def scan_operator(
    *,
    axis: common.Dimension,
    forward: bool,
    init: core_defs.Scalar,
    backend: next_backend.Backend | eve.NothingType | None,
    grid_type: common.GridType | None,
) -> Callable[[types.FunctionType], FieldOperator[foast.ScanOperator]]: ...


def scan_operator(
    definition: Optional[types.FunctionType] = None,
    *,
    axis: common.Dimension,
    forward: bool = True,
    init: core_defs.Scalar = 0.0,
    backend: next_backend.Backend | None | eve.NothingType = eve.NOTHING,
    grid_type: common.GridType | None = None,
) -> (
    FieldOperator[foast.ScanOperator]
    | Callable[[types.FunctionType], FieldOperator[foast.ScanOperator]]
):
    """
    Generate an implementation of the scan operator from a Python function object.

    Arguments:
        definition: Function from scalars to a scalar.

    Keyword Arguments:
        axis: A :ref:`Dimension` to reduce over.
        forward: Boolean specifying the direction.
        init: Initial value for the carry argument of the scan pass.

    Examples:
        >>> import numpy as np
        >>> import gt4py.next as gtx
        >>> from gt4py.next.iterator import embedded
        >>> embedded._column_range = 1  # implementation detail
        >>> KDim = gtx.Dimension("K", kind=gtx.DimensionKind.VERTICAL)
        >>> inp = gtx.as_field([KDim], np.ones((10,)))
        >>> out = gtx.as_field([KDim], np.zeros((10,)))
        >>> @gtx.scan_operator(axis=KDim, forward=True, init=0.0)
        ... def scan_operator(carry: float, val: float) -> float:
        ...     return carry + val
        >>> scan_operator(inp, out=out, offset_provider={})  # doctest: +SKIP
        >>> out.array()  # doctest: +SKIP
        array([ 1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9., 10.])
    """
    # TODO(tehrengruber): enable doctests again. For unknown / obscure reasons
    #  the above doctest fails when executed using `pytest --doctest-modules`.

    def scan_operator_inner(definition: types.FunctionType) -> FieldOperator:
        return FieldOperator.from_function(
            definition,
            typing.cast(
                next_backend.Backend | None, DEFAULT_BACKEND if backend is eve.NOTHING else backend
            ),
            grid_type,
            operator_node_cls=foast.ScanOperator,
            operator_attributes={"axis": axis, "forward": forward, "init": init},
        )

    return scan_operator_inner if definition is None else scan_operator_inner(definition)


@ffront_stages.add_content_to_fingerprint.register
def add_fieldop_to_fingerprint(obj: FieldOperator, hasher: xtyping.HashlibAlgorithm) -> None:
    ffront_stages.add_content_to_fingerprint(obj.definition_stage, hasher)
    ffront_stages.add_content_to_fingerprint(obj.backend, hasher)


@ffront_stages.add_content_to_fingerprint.register
def add_foast_fieldop_to_fingerprint(
    obj: FieldOperatorFromFoast, hasher: xtyping.HashlibAlgorithm
) -> None:
    ffront_stages.add_content_to_fingerprint(obj.foast_stage, hasher)
    ffront_stages.add_content_to_fingerprint(obj.backend, hasher)


@ffront_stages.add_content_to_fingerprint.register
def add_program_to_fingerprint(obj: Program, hasher: xtyping.HashlibAlgorithm) -> None:
    ffront_stages.add_content_to_fingerprint(obj.definition_stage, hasher)
    ffront_stages.add_content_to_fingerprint(obj.backend, hasher)


@ffront_stages.add_content_to_fingerprint.register
def add_past_program_to_fingerprint(obj: ProgramFromPast, hasher: xtyping.HashlibAlgorithm) -> None:
    ffront_stages.add_content_to_fingerprint(obj.past_stage, hasher)
    ffront_stages.add_content_to_fingerprint(obj.backend, hasher)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import collections
from typing import Any, Iterable, Optional

from gt4py.next import common
from gt4py.next.ffront import fbuiltins
from gt4py.next.ffront.gtcallable import GTCallable


def _get_closure_vars_recursively(closure_vars: dict[str, Any]) -> dict[str, Any]:
    all_closure_vars = collections.ChainMap(closure_vars)

    for closure_var in closure_vars.values():
        if isinstance(closure_var, GTCallable):
            # if the closure ref has closure refs by itself, also add them
            if child_closure_vars := closure_var.__gt_closure_vars__():
                all_child_closure_vars = _get_closure_vars_recursively(child_closure_vars)

                collisions: list[str] = []
                for potential_collision in set(closure_vars) & set(all_child_closure_vars):
                    if (
                        closure_vars[potential_collision]
                        != all_child_closure_vars[potential_collision]
                    ):
                        collisions.append(potential_collision)
                if collisions:
                    raise NotImplementedError(
                        f"Using closure vars with same name but different value "
                        f"across functions is not implemented yet. \n"
                        f"Collisions: '{',  '.join(collisions)}'."
                    )

                all_closure_vars = collections.ChainMap(all_closure_vars, all_child_closure_vars)
    return dict(all_closure_vars)


def _filter_closure_vars_by_type(closure_vars: dict[str, Any], *types: type) -> dict[str, Any]:
    return {name: value for name, value in closure_vars.items() if isinstance(value, types)}


def _deduce_grid_type(
    requested_grid_type: Optional[common.GridType],
    offsets_and_dimensions: Iterable[fbuiltins.FieldOffset | common.Dimension],
) -> common.GridType:
    """
    Derive grid type from actually occurring dimensions and check against optional user request.

    Unstructured grid type is consistent with any kind of offset, cartesian
    is easier to optimize for but only allowed in the absence of unstructured
    dimensions and offsets.
    """

    def is_cartesian_offset(o: fbuiltins.FieldOffset) -> bool:
        return len(o.target) == 1 and o.source == o.target[0]

    deduced_grid_type = common.GridType.CARTESIAN
    for o in offsets_and_dimensions:
        if isinstance(o, fbuiltins.FieldOffset) and not is_cartesian_offset(o):
            deduced_grid_type = common.GridType.UNSTRUCTURED
            break
        if isinstance(o, common.Dimension) and o.kind == common.DimensionKind.LOCAL:
            deduced_grid_type = common.GridType.UNSTRUCTURED
            break

    if (
        requested_grid_type == common.GridType.CARTESIAN
        and deduced_grid_type == common.GridType.UNSTRUCTURED
    ):
        raise ValueError(
            "'grid_type == GridType.CARTESIAN' was requested, but unstructured 'FieldOffset' or local 'Dimension' was found."
        )

    return deduced_grid_type if requested_grid_type is None else requested_grid_type



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import math

from gt4py.eve import NodeTranslator
from gt4py.next.iterator import ir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.iterator.transforms.inline_lambdas import InlineLambdas


def _is_power_call(node: ir.FunCall) -> bool:
    """Match expressions of the form `power(base, integral_literal)`."""
    return (
        isinstance(node.fun, ir.SymRef)
        and node.fun.id == "power"
        and isinstance(node.args[1], ir.Literal)
        and float(node.args[1].value) == int(node.args[1].value)
        and node.args[1].value >= im.literal_from_value(0).value
    )


def _compute_integer_power_of_two(exp: int) -> int:
    return math.floor(math.log2(exp))


@dataclasses.dataclass
class PowerUnrolling(NodeTranslator):
    max_unroll: int

    @classmethod
    def apply(cls, node: ir.Node, max_unroll: int = 5) -> ir.Node:
        return cls(max_unroll=max_unroll).visit(node)

    def visit_FunCall(self, node: ir.FunCall) -> ir.Expr:
        new_node = self.generic_visit(node)

        if _is_power_call(new_node):
            assert len(new_node.args) == 2
            # Check if unroll should be performed or if exponent is too large
            base, exponent = new_node.args[0], int(new_node.args[1].value)
            if 1 <= exponent <= self.max_unroll:
                # Calculate and store powers of two of the base as long as they are smaller than the exponent.
                # Do the same (using the stored values) with the remainder and multiply computed values.
                pow_cur = _compute_integer_power_of_two(exponent)
                pow_max = pow_cur
                remainder = exponent

                # Build target expression
                ret: ir.Expr = im.ref(f"power_{2**pow_max}")
                remainder -= 2**pow_cur
                while remainder > 0:
                    pow_cur = _compute_integer_power_of_two(remainder)
                    remainder -= 2**pow_cur

                    ret = im.multiplies_(ret, f"power_{2**pow_cur}")

                # Nest target expression to avoid multiple redundant evaluations
                for i in range(pow_max, 0, -1):
                    ret = im.let(
                        f"power_{2**i}",
                        im.multiplies_(f"power_{2 ** (i - 1)}", f"power_{2 ** (i - 1)}"),
                    )(ret)
                ret = im.let("power_1", base)(ret)

                # Simplify expression in case of SymRef by resolving let statements
                if isinstance(base, ir.SymRef):
                    return InlineLambdas.apply(ret, opcount_preserving=True)
                else:
                    return ret
        return new_node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import types

import numpy as np

from gt4py._core import definitions as core_defs


try:
    import cupy
except ImportError:
    cupy = None


def array_namespace(array: core_defs.NDArrayObject) -> types.ModuleType:
    """
    Get the namespace of the array.

    This is defined in https://data-apis.org/array-api/latest/API_specification/generated/array_api.array.__array_namespace__.html,
    however not implemented in CuPy < 14.
    """
    if hasattr(array, "__array_namespace__"):
        return array.__array_namespace__()
    else:
        if isinstance(array, np.ndarray):
            return np
        if cupy is not None and isinstance(array, cupy.ndarray):
            return cupy
        raise TypeError(f"Could not determine array namespace of {array} of type {type(array)}")



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, cast

import gt4py.next.ffront.field_operator_ast as foast
from gt4py.eve import NodeTranslator, traits
from gt4py.eve.concepts import SourceLocation, SymbolName, SymbolRef
from gt4py.next.ffront import dialect_ast_enums
from gt4py.next.ffront.fbuiltins import TYPE_BUILTIN_NAMES
from gt4py.next.type_system import type_specifications as ts
from gt4py.next.type_system.type_translation import from_type_hint


@dataclass
class TypeAliasReplacement(NodeTranslator, traits.VisitorWithSymbolTableTrait):
    """
    Replace Type Aliases with their actual type.

    After this pass, the type aliases used for explicit construction of literal
    values and for casting field values are replaced by their actual types.
    """

    closure_vars: dict[str, Any]

    @classmethod
    def apply(
        cls, node: foast.FunctionDefinition | foast.FieldOperator, closure_vars: dict[str, Any]
    ) -> tuple[foast.FunctionDefinition, dict[str, Any]]:
        foast_node = cls(closure_vars=closure_vars).visit(node)
        new_closure_vars = closure_vars.copy()
        for key, value in closure_vars.items():
            if isinstance(value, type) and key not in TYPE_BUILTIN_NAMES:
                new_closure_vars[value.__name__] = closure_vars[key]
        return foast_node, new_closure_vars

    def is_type_alias(self, node_id: SymbolName | SymbolRef) -> bool:
        return (
            node_id in self.closure_vars
            and isinstance(self.closure_vars[node_id], type)
            and node_id not in TYPE_BUILTIN_NAMES
        )

    def visit_Name(self, node: foast.Name, **kwargs: Any) -> foast.Name:
        if self.is_type_alias(node.id):
            return foast.Name(
                id=self.closure_vars[node.id].__name__, location=node.location, type=node.type
            )
        return node

    def _update_closure_var_symbols(
        self, closure_vars: list[foast.Symbol], location: SourceLocation
    ) -> list[foast.Symbol]:
        new_closure_vars: list[foast.Symbol] = []
        existing_type_names: set[str] = set()

        for var in closure_vars:
            if self.is_type_alias(var.id):
                actual_type_name = self.closure_vars[var.id].__name__
                # Avoid multiple definitions of a type in closure_vars
                if actual_type_name not in existing_type_names:
                    new_closure_vars.append(
                        foast.Symbol(
                            id=actual_type_name,
                            type=ts.FunctionType(
                                pos_or_kw_args={},
                                kw_only_args={},
                                pos_only_args=[ts.DeferredType(constraint=ts.ScalarType)],
                                returns=cast(
                                    ts.DataType, from_type_hint(self.closure_vars[var.id])
                                ),
                            ),
                            namespace=dialect_ast_enums.Namespace.CLOSURE,
                            location=location,
                        )
                    )
                    existing_type_names.add(actual_type_name)
            elif var.id not in existing_type_names:
                new_closure_vars.append(var)
                existing_type_names.add(var.id)

        return new_closure_vars

    def visit_FunctionDefinition(
        self, node: foast.FunctionDefinition, **kwargs: Any
    ) -> foast.FunctionDefinition:
        return foast.FunctionDefinition(
            id=node.id,
            params=node.params,
            body=self.visit(node.body, **kwargs),
            closure_vars=self._update_closure_var_symbols(node.closure_vars, node.location),
            location=node.location,
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Iterable, Optional, Protocol

import dace
from dace import subsets as dace_subsets

from gt4py.next import common as gtx_common, utils as gtx_utils
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.program_processors.runners.dace import (
    gtir_dataflow,
    gtir_domain,
    gtir_python_codegen,
    gtir_to_sdfg,
    gtir_to_sdfg_types,
    gtir_to_sdfg_utils,
    utils as gtx_dace_utils,
)
from gt4py.next.program_processors.runners.dace.gtir_to_sdfg_concat_where import (
    translate_concat_where,
)
from gt4py.next.program_processors.runners.dace.gtir_to_sdfg_scan import translate_scan
from gt4py.next.type_system import type_info as ti, type_specifications as ts


if TYPE_CHECKING:
    from gt4py.next.program_processors.runners.dace import gtir_to_sdfg


class PrimitiveTranslator(Protocol):
    @abc.abstractmethod
    def __call__(
        self,
        node: gtir.Node,
        ctx: gtir_to_sdfg.SubgraphContext,
        sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    ) -> gtir_to_sdfg_types.FieldopResult:
        """Creates the dataflow subgraph representing a GTIR primitive function.

        This method is used by derived classes to build a specialized subgraph
        for a specific GTIR primitive function.

        Args:
            node: The GTIR node describing the primitive to be lowered.
            ctx: The SDFG context in which to lower the primitive subgraph.
            sdfg_builder: The object responsible for visiting child nodes of the primitive node.

        Returns:
            A list of data access nodes and the associated GT4Py data type, which provide
            access to the result of the primitive subgraph. The GT4Py data type is useful
            in the case the returned data is an array, because the type provdes the domain
            information (e.g. order of dimensions, dimension types).
        """


def _parse_fieldop_arg(
    node: gtir.Expr,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    domain: gtir_domain.FieldopDomain,
) -> (
    gtir_dataflow.IteratorExpr
    | gtir_dataflow.MemletExpr
    | tuple[gtir_dataflow.IteratorExpr | gtir_dataflow.MemletExpr | tuple[Any, ...], ...]
):
    """Helper method to visit an expression passed as argument to a field operator."""

    arg = sdfg_builder.visit(node, ctx=ctx)

    if isinstance(arg, gtir_to_sdfg_types.FieldopData):
        return arg.get_local_view(domain, ctx.sdfg)
    else:
        # handle tuples of fields
        return gtx_utils.tree_map(lambda targ: targ.get_local_view(domain))(arg)


def _create_field_operator_impl(
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    field_domain: gtir_domain.FieldopDomain,
    output_edge: gtir_dataflow.DataflowOutputEdge,
    output_type: ts.FieldType,
    map_exit: dace.nodes.MapExit,
) -> gtir_to_sdfg_types.FieldopData:
    """
    Helper method to allocate a temporary array that stores one field computed
    by a field operator.

    This method is called by `_create_field_operator()`.

    Args:
        ctx: The SDFG context in which to lower the field operator.
        sdfg_builder: The object used to build the map scope in the provided SDFG.
        domain: The domain of the field operator that computes the field.
        output_edge: The dataflow write edge representing the output data.
        output_type: The GT4Py field type descriptor.
        map_exit: The `MapExit` node of the field operator map scope.

    Returns:
        The field data descriptor, which includes the field access node in the
        given `state` and the field domain offset.
    """
    dataflow_output_desc = output_edge.result.dc_node.desc(ctx.sdfg)

    # the memory layout of the output field follows the field operator compute domain
    field_dims, field_origin, field_shape = gtir_domain.get_field_layout(
        field_domain, ctx.target_domain
    )
    if len(field_domain) == 0:
        # The field operator computes a zero-dimensional field, and the data subset
        # is set later depending on the element type (`ts.ListType` or `ts.ScalarType`)
        field_subset = dace_subsets.Range([])
    else:
        field_indices = gtir_domain.get_domain_indices(field_dims, field_origin)
        field_subset = dace_subsets.Range.from_indices(field_indices)

    if isinstance(output_edge.result.gt_dtype, ts.ScalarType):
        if output_edge.result.gt_dtype != output_type.dtype:
            raise TypeError(
                f"Type mismatch, expected {output_type.dtype} got {output_edge.result.gt_dtype}."
            )
        assert isinstance(dataflow_output_desc, dace.data.Scalar)
    else:
        assert isinstance(output_type.dtype, ts.ListType)
        assert isinstance(output_edge.result.gt_dtype.element_type, ts.ScalarType)
        if output_edge.result.gt_dtype.element_type != output_type.dtype.element_type:
            raise TypeError(
                f"Type mismatch, expected {output_type.dtype.element_type} got {output_edge.result.gt_dtype.element_type}."
            )
        assert isinstance(dataflow_output_desc, dace.data.Array)
        assert len(dataflow_output_desc.shape) == 1
        # extend the array with the local dimensions added by the field operator (e.g. `neighbors`)
        assert output_edge.result.gt_dtype.offset_type is not None
        field_shape = [*field_shape, dataflow_output_desc.shape[0]]
        field_subset = field_subset + dace_subsets.Range.from_array(dataflow_output_desc)

    # allocate local temporary storage
    if len(field_shape) == 0:  # zero-dimensional field
        field_name, _ = sdfg_builder.add_temp_scalar(ctx.sdfg, dataflow_output_desc.dtype)
        field_subset = dace_subsets.Range.from_string("0")
    else:
        field_name, _ = sdfg_builder.add_temp_array(
            ctx.sdfg, field_shape, dataflow_output_desc.dtype
        )
    field_node = ctx.state.add_access(field_name)

    # and here the edge writing the dataflow result data through the map exit node
    output_edge.connect(map_exit, field_node, field_subset)

    return gtir_to_sdfg_types.FieldopData(
        field_node, ts.FieldType(field_dims, output_edge.result.gt_dtype), tuple(field_origin)
    )


def _create_field_operator(
    ctx: gtir_to_sdfg.SubgraphContext,
    domain: gtir_domain.FieldopDomain,
    node_type: ts.FieldType | ts.TupleType,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    input_edges: Iterable[gtir_dataflow.DataflowInputEdge],
    output_tree: tuple[gtir_dataflow.DataflowOutputEdge | tuple[Any, ...], ...],
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Helper method to build the output of a field operator, which can consist of
    a single field or a tuple of fields.

    A tuple of fields is returned when one stencil computes a grid point on multiple
    fields: for each field, this method will call `_create_field_operator_impl()`.

    Args:
        ctx: The SDFG context in which to lower the field operator.
        domain: The domain of the field operator that computes the field.
        node_type: The GT4Py type of the IR node that produces this field.
        sdfg_builder: The object used to build the map scope in the provided SDFG.
        input_edges: List of edges to pass input data into the dataflow.
        output_tree: A tree representation of the dataflow output data.

    Returns:
        The descriptor of the field operator result, which can be either a single
        field or a tuple fields.
    """

    if len(domain) == 0:
        # create a trivial map for zero-dimensional fields
        map_range = {
            "__gt4py_zerodim": "0",
        }
    else:
        # create map range corresponding to the field operator domain
        map_range = {
            gtir_to_sdfg_utils.get_map_variable(
                domain_range.dim
            ): f"{domain_range.start}:{domain_range.stop}"
            for domain_range in domain
        }
    map_entry, map_exit = sdfg_builder.add_map("fieldop", ctx.state, map_range)

    # here we setup the edges passing through the map entry node
    for edge in input_edges:
        edge.connect(map_entry)

    if isinstance(node_type, ts.FieldType):
        assert len(output_tree) == 1 and isinstance(
            output_tree[0], gtir_dataflow.DataflowOutputEdge
        )
        output_edge = output_tree[0]
        return _create_field_operator_impl(
            ctx, sdfg_builder, domain, output_edge, node_type, map_exit
        )
    else:
        # handle tuples of fields
        output_symbol_tree = gtir_to_sdfg_utils.make_symbol_tree("x", node_type)
        return gtx_utils.tree_map(
            lambda output_edge, output_sym: _create_field_operator_impl(
                ctx, sdfg_builder, domain, output_edge, output_sym.type, map_exit
            )
        )(output_tree, output_symbol_tree)


def translate_as_fieldop(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Generates the dataflow subgraph for the `as_fieldop` builtin function.

    Expects a `FunCall` node with two arguments:
    1. a lambda function representing the stencil, which is lowered to a dataflow subgraph
    2. the domain of the field operator, which is used as map range

    The dataflow can be as simple as a single tasklet, or implement a local computation
    as a composition of tasklets and even include a map to range on local dimensions (e.g.
    neighbors and map builtins).
    The stencil dataflow is instantiated inside a map scope, which applies the stencil
    over the field domain.
    """
    assert isinstance(node, gtir.FunCall)
    assert cpm.is_call_to(node.fun, "as_fieldop")
    assert isinstance(node.type, (ts.FieldType, ts.TupleType))

    fun_node = node.fun
    assert len(fun_node.args) == 2
    fieldop_expr, domain_expr = fun_node.args

    if cpm.is_call_to(fieldop_expr, "scan"):
        return translate_scan(node, ctx, sdfg_builder)

    if cpm.is_ref_to(fieldop_expr, "deref"):
        # Special usage of 'deref' as argument to fieldop expression, to pass a scalar
        # value to 'as_fieldop' function. It results in broadcasting the scalar value
        # over the field domain.
        assert isinstance(node.type, ts.FieldType)
        stencil_expr = im.lambda_("a")(im.deref("a"))
        stencil_expr.expr.type = node.type.dtype
    elif isinstance(fieldop_expr, gtir.Lambda):
        # Default case, handled below: the argument expression is a lambda function
        # representing the stencil operation to be computed over the field domain.
        stencil_expr = fieldop_expr
    else:
        raise NotImplementedError(
            f"Expression type '{type(fieldop_expr)}' not supported as argument to 'as_fieldop' node."
        )

    # parse the domain of the field operator
    domain = gtir_domain.extract_domain(domain_expr)

    # visit the list of arguments to be passed to the lambda expression
    fieldop_args = [_parse_fieldop_arg(arg, ctx, sdfg_builder, domain) for arg in node.args]

    # represent the field operator as a mapped tasklet graph, which will range over the field domain
    input_edges, output_edges = gtir_dataflow.translate_lambda_to_dataflow(
        ctx.sdfg, ctx.state, sdfg_builder, stencil_expr, fieldop_args
    )

    return _create_field_operator(ctx, domain, node.type, sdfg_builder, input_edges, output_edges)


def _construct_if_branch_output(
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    field_domain: gtir.Expr,
    sym: gtir.Sym,
    true_br: gtir_to_sdfg_types.FieldopData,
    false_br: gtir_to_sdfg_types.FieldopData,
) -> gtir_to_sdfg_types.FieldopData:
    """
    Helper function called by `translate_if()` to allocate a temporary field to store
    the result of an if expression.
    """
    assert true_br.gt_type == false_br.gt_type
    out_type = true_br.gt_type

    if isinstance(sym.type, ts.ScalarType):
        assert sym.type == out_type
        dtype = gtx_dace_utils.as_dace_type(sym.type)
        out, _ = sdfg_builder.add_temp_scalar(ctx.sdfg, dtype)
        out_node = ctx.state.add_access(out)
        return gtir_to_sdfg_types.FieldopData(out_node, sym.type, origin=())

    assert isinstance(out_type, ts.FieldType)
    assert isinstance(sym.type, ts.FieldType)
    dims, origin, shape = gtir_domain.get_field_layout(
        gtir_domain.extract_domain(field_domain), ctx.target_domain
    )
    assert dims == out_type.dims

    if isinstance(out_type.dtype, ts.ScalarType):
        dtype = gtx_dace_utils.as_dace_type(out_type.dtype)
    else:
        assert isinstance(out_type.dtype, ts.ListType)
        assert out_type.dtype.offset_type is not None
        assert isinstance(out_type.dtype.element_type, ts.ScalarType)
        dtype = gtx_dace_utils.as_dace_type(out_type.dtype.element_type)
        offset_provider_type = sdfg_builder.get_offset_provider_type(
            out_type.dtype.offset_type.value
        )
        assert isinstance(offset_provider_type, gtx_common.NeighborConnectivityType)
        shape = [*shape, offset_provider_type.max_neighbors]

    out, _ = sdfg_builder.add_temp_array(ctx.sdfg, shape, dtype)
    out_node = ctx.state.add_access(out)

    return gtir_to_sdfg_types.FieldopData(out_node, out_type, tuple(origin))


def _write_if_branch_output(
    ctx: gtir_to_sdfg.SubgraphContext,
    src: gtir_to_sdfg_types.FieldopData,
    dst: gtir_to_sdfg_types.FieldopData,
) -> None:
    """
    Helper function called by `translate_if()` to write the result of an if-branch,
    here `src` field, to the output 'dst' field. The data subset is based on the
    domain of the `dst` field. Therefore, the full shape of `dst` array is written.
    """
    if src.gt_type != dst.gt_type:
        raise ValueError(
            f"Source and destination type mismatch, '{dst.gt_type}' vs '{src.gt_type}'."
        )
    dst_node = ctx.state.add_access(dst.dc_node.data)
    dst_shape = dst_node.desc(ctx.sdfg).shape

    if isinstance(src.gt_type, ts.ScalarType):
        ctx.state.add_nedge(
            src.dc_node,
            dst_node,
            dace.Memlet(data=src.dc_node.data, subset="0"),
        )
    else:
        if isinstance(src.gt_type.dtype, ts.ListType):
            src_origin = [*src.origin, 0]
            dst_origin = [*dst.origin, 0]
        else:
            src_origin = [*src.origin]
            dst_origin = [*dst.origin]

        data_subset = dace_subsets.Range(
            (
                f"{dst_start - src_start}",
                f"{dst_start - src_start + size - 1}",  # subtract 1 because the range boundaries are included
                1,
            )
            for src_start, dst_start, size in zip(src_origin, dst_origin, dst_shape, strict=True)
        )

        ctx.state.add_nedge(
            src.dc_node,
            dst_node,
            dace.Memlet(
                data=src.dc_node.data,
                subset=data_subset,
                other_subset=dace_subsets.Range.from_array(dst_node.desc(ctx.sdfg)),
            ),
        )


def translate_if(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """Generates the dataflow subgraph for the `if_` builtin function."""
    assert cpm.is_call_to(node, "if_")
    assert len(node.args) == 3
    cond_expr, true_expr, false_expr = node.args

    # expect condition as first argument
    if_stmt = gtir_python_codegen.get_source(cond_expr)

    # evaluate the if-condition in a new entry state and use the current head state
    # to join the true/false branch states as follows:
    #
    #               ------------
    #           === |   cond   | ===
    #          ||   ------------   ||
    #          \/                  \/
    #     ------------       -------------
    #     |   true   |       |   false   |
    #     ------------       -------------
    #          ||                  ||
    #          ||   ------------   ||
    #           ==> |   head   | <==
    #               ------------
    #
    cond_state = ctx.sdfg.add_state_before(ctx.state, ctx.state.label + "_cond")
    ctx.sdfg.remove_edge(ctx.sdfg.out_edges(cond_state)[0])

    # expect true branch as second argument
    true_state = ctx.sdfg.add_state(ctx.state.label + "_true_branch")
    tbranch_ctx = gtir_to_sdfg.SubgraphContext(ctx.sdfg, true_state, ctx.target_domain)
    ctx.sdfg.add_edge(cond_state, true_state, dace.InterstateEdge(condition=if_stmt))
    ctx.sdfg.add_edge(true_state, ctx.state, dace.InterstateEdge())

    # and false branch as third argument
    false_state = ctx.sdfg.add_state(ctx.state.label + "_false_branch")
    fbranch_ctx = gtir_to_sdfg.SubgraphContext(ctx.sdfg, false_state, ctx.target_domain)
    ctx.sdfg.add_edge(cond_state, false_state, dace.InterstateEdge(condition=f"not({if_stmt})"))
    ctx.sdfg.add_edge(false_state, ctx.state, dace.InterstateEdge())

    true_br_result = sdfg_builder.visit(true_expr, ctx=tbranch_ctx)
    false_br_result = sdfg_builder.visit(false_expr, ctx=fbranch_ctx)

    if isinstance(node.type, ts.TupleType):
        symbol_tree = gtir_to_sdfg_utils.make_symbol_tree("x", node.type)
        if isinstance(node.annex.domain, tuple):
            domain_tree = node.annex.domain
        else:
            # TODO(edopao): this is a workaround for some IR nodes where the inferred
            #   domain on a tuple of fields is not a tuple, see `test_execution.py::test_ternary_operator_tuple()`
            domain_tree = gtx_utils.tree_map(lambda _: node.annex.domain)(symbol_tree)
        node_output = gtx_utils.tree_map(
            lambda sym,
            domain,
            true_br,
            false_br,
            _ctx=ctx,
            sdfg_builder=sdfg_builder: _construct_if_branch_output(
                _ctx,
                sdfg_builder,
                domain,
                sym,
                true_br,
                false_br,
            )
        )(
            symbol_tree,
            domain_tree,
            true_br_result,
            false_br_result,
        )
        gtx_utils.tree_map(
            lambda src, dst, _ctx=tbranch_ctx: _write_if_branch_output(_ctx, src, dst)
        )(true_br_result, node_output)
        gtx_utils.tree_map(
            lambda src, dst, _ctx=fbranch_ctx: _write_if_branch_output(_ctx, src, dst)
        )(false_br_result, node_output)
    else:
        node_output = _construct_if_branch_output(
            ctx,
            sdfg_builder,
            node.annex.domain,
            im.sym("x", node.type),
            true_br_result,
            false_br_result,
        )
        _write_if_branch_output(tbranch_ctx, true_br_result, node_output)
        _write_if_branch_output(fbranch_ctx, false_br_result, node_output)

    return node_output


def translate_index(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """
    Lowers the `index` builtin function to a mapped tasklet that writes the dimension
    index values to a transient array. The extent of the index range is taken from
    the domain information that should be present in the node annex.
    """
    assert cpm.is_call_to(node, "index")
    assert isinstance(node.type, ts.FieldType)

    assert "domain" in node.annex
    domain = gtir_domain.extract_domain(node.annex.domain)
    assert len(domain) == 1
    dim_index = gtir_to_sdfg_utils.get_map_variable(domain[0].dim)

    index_data, _ = sdfg_builder.add_temp_scalar(ctx.sdfg, gtir_to_sdfg_types.INDEX_DTYPE)
    index_node = ctx.state.add_access(index_data)
    index_value = gtir_dataflow.ValueExpr(
        dc_node=index_node,
        gt_dtype=gtx_dace_utils.as_itir_type(gtir_to_sdfg_types.INDEX_DTYPE),
    )
    index_write_tasklet = sdfg_builder.add_tasklet(
        "index",
        ctx.state,
        inputs={},
        outputs={"__val"},
        code=f"__val = {dim_index}",
    )
    ctx.state.add_edge(
        index_write_tasklet,
        "__val",
        index_node,
        None,
        dace.Memlet(data=index_data, subset="0"),
    )

    input_edges = [
        gtir_dataflow.EmptyInputEdge(ctx.state, index_write_tasklet),
    ]
    output_edge = gtir_dataflow.DataflowOutputEdge(ctx.state, index_value)
    return _create_field_operator(ctx, domain, node.type, sdfg_builder, input_edges, (output_edge,))


def _get_data_nodes(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    data_name: str,
    data_type: ts.DataType,
) -> gtir_to_sdfg_types.FieldopResult:
    if isinstance(data_type, ts.FieldType):
        data_node = state.add_access(data_name)
        return sdfg_builder.make_field(data_node, data_type)

    elif isinstance(data_type, ts.ScalarType):
        if data_name in sdfg.symbols:
            data_node = _get_symbolic_value(
                sdfg, state, sdfg_builder, data_name, data_type, temp_name=f"__{data_name}"
            )
        else:
            data_node = state.add_access(data_name)
        return gtir_to_sdfg_types.FieldopData(data_node, data_type, origin=())

    elif isinstance(data_type, ts.TupleType):
        symbol_tree = gtir_to_sdfg_utils.make_symbol_tree(data_name, data_type)
        return gtx_utils.tree_map(
            lambda sym: _get_data_nodes(sdfg, state, sdfg_builder, sym.id, sym.type)
        )(symbol_tree)

    else:
        raise NotImplementedError(f"Symbol type {type(data_type)} not supported.")


def _get_symbolic_value(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
    symbolic_expr: dace.symbolic.SymExpr,
    scalar_type: ts.ScalarType,
    temp_name: Optional[str] = None,
) -> dace.nodes.AccessNode:
    tasklet_node = sdfg_builder.add_tasklet(
        "get_value",
        state,
        {},
        {"__out"},
        f"__out = {symbolic_expr}",
    )
    temp_name, _ = sdfg.add_scalar(
        temp_name or sdfg.temp_data_name(),
        gtx_dace_utils.as_dace_type(scalar_type),
        find_new_name=True,
        transient=True,
    )
    data_node = state.add_access(temp_name)
    state.add_edge(
        tasklet_node,
        "__out",
        data_node,
        None,
        dace.Memlet(data=temp_name, subset="0"),
    )
    return data_node


def translate_literal(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """Generates the dataflow subgraph for a `ir.Literal` node."""
    assert isinstance(node, gtir.Literal)

    data_type = node.type
    data_node = _get_symbolic_value(ctx.sdfg, ctx.state, sdfg_builder, node.value, data_type)

    return gtir_to_sdfg_types.FieldopData(data_node, data_type, origin=())


def translate_make_tuple(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    assert cpm.is_call_to(node, "make_tuple")
    return tuple(sdfg_builder.visit(arg, ctx=ctx) for arg in node.args)


def translate_tuple_get(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    assert cpm.is_call_to(node, "tuple_get")
    assert len(node.args) == 2

    if not isinstance(node.args[0], gtir.Literal):
        raise ValueError("Tuple can only be subscripted with compile-time constants.")
    assert ti.is_integral(node.args[0].type)
    index = int(node.args[0].value)

    data_nodes = sdfg_builder.visit(node.args[1], ctx=ctx)
    if isinstance(data_nodes, gtir_to_sdfg_types.FieldopData):
        raise ValueError(f"Invalid tuple expression {node}")
    unused_arg_nodes: Iterable[gtir_to_sdfg_types.FieldopData] = gtx_utils.flatten_nested_tuple(
        tuple(arg for i, arg in enumerate(data_nodes) if i != index)
    )
    ctx.state.remove_nodes_from(
        [arg.dc_node for arg in unused_arg_nodes if ctx.state.degree(arg.dc_node) == 0]
    )
    return data_nodes[index]


def translate_scalar_expr(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    assert isinstance(node, gtir.FunCall)
    assert isinstance(node.type, ts.ScalarType)

    args = []
    connectors = []
    scalar_expr_args = []

    for i, arg_expr in enumerate(node.args):
        visit_expr = True
        if isinstance(arg_expr, gtir.SymRef):
            try:
                # check if symbol is defined in the GT4Py program, throws `KeyError` exception if undefined
                sdfg_builder.get_symbol_type(arg_expr.id)
            except KeyError:
                # all `SymRef` should refer to symbols defined in the program, except in case of non-variable argument,
                # e.g. the type name `float64` used in casting expressions like `cast_(variable, float64)`
                visit_expr = False

        if visit_expr:
            # we visit the argument expression and obtain the access node to
            # a scalar data container, which will be connected to the tasklet
            arg = sdfg_builder.visit(arg_expr, ctx=ctx)
            if not (
                isinstance(arg, gtir_to_sdfg_types.FieldopData)
                and isinstance(node.type, ts.ScalarType)
            ):
                raise ValueError(f"Invalid argument to scalar expression {arg_expr}.")
            param = f"__arg{i}"
            args.append(arg.dc_node)
            connectors.append(param)
            scalar_expr_args.append(gtir.SymRef(id=param))
        else:
            assert isinstance(arg_expr, gtir.SymRef)
            scalar_expr_args.append(arg_expr)

    # we visit the scalar expression replacing the input arguments with the corresponding data connectors
    scalar_node = gtir.FunCall(fun=node.fun, args=scalar_expr_args)
    python_code = gtir_python_codegen.get_source(scalar_node)
    tasklet_node = sdfg_builder.add_tasklet(
        name="scalar_expr",
        state=ctx.state,
        inputs=set(connectors),
        outputs={"__out"},
        code=f"__out = {python_code}",
    )
    # create edges for the input data connectors
    for arg_node, conn in zip(args, connectors, strict=True):
        ctx.state.add_edge(
            arg_node,
            None,
            tasklet_node,
            conn,
            dace.Memlet(data=arg_node.data, subset="0"),
        )
    # finally, create temporary for the result value
    temp_name, _ = sdfg_builder.add_temp_scalar(ctx.sdfg, gtx_dace_utils.as_dace_type(node.type))
    temp_node = ctx.state.add_access(temp_name)
    ctx.state.add_edge(
        tasklet_node,
        "__out",
        temp_node,
        None,
        dace.Memlet(data=temp_name, subset="0"),
    )

    return gtir_to_sdfg_types.FieldopData(temp_node, node.type, origin=())


def translate_symbol_ref(
    node: gtir.Node,
    ctx: gtir_to_sdfg.SubgraphContext,
    sdfg_builder: gtir_to_sdfg.SDFGBuilder,
) -> gtir_to_sdfg_types.FieldopResult:
    """Generates the dataflow subgraph for a `ir.SymRef` node."""
    assert isinstance(node, gtir.SymRef)

    symbol_name = str(node.id)
    # we retrieve the type of the symbol in the GT4Py prgram
    gt_symbol_type = sdfg_builder.get_symbol_type(symbol_name)

    # Create new access node in current state. It is possible that multiple
    # access nodes are created in one state for the same data container.
    # We rely on the dace simplify pass to remove duplicated access nodes.
    return _get_data_nodes(ctx.sdfg, ctx.state, sdfg_builder, symbol_name, gt_symbol_type)


if TYPE_CHECKING:
    # Use type-checking to assert that all translator functions implement the `PrimitiveTranslator` protocol
    __primitive_translators: list[PrimitiveTranslator] = [
        translate_as_fieldop,
        translate_concat_where,
        translate_if,
        translate_index,
        translate_literal,
        translate_make_tuple,
        translate_tuple_get,
        translate_scalar_expr,
        translate_scan,
        translate_symbol_ref,
    ]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import concurrent.futures
import dataclasses
import functools
import itertools
from collections.abc import Sequence
from typing import Any, TypeAlias

from gt4py._core import definitions as core_defs
from gt4py.eve import extended_typing, utils as eve_utils
from gt4py.next import backend as gtx_backend, common, config, errors
from gt4py.next.ffront import stages as ffront_stages, type_specifications as ts_ffront
from gt4py.next.otf import arguments, stages
from gt4py.next.type_system import type_info, type_specifications as ts, type_translation as tt


# TODO(havogt): We would like this to be a ProcessPoolExecutor, which requires (to decide what) to pickle.
_async_compilation_pool = concurrent.futures.ThreadPoolExecutor(max_workers=config.BUILD_JOBS)

ScalarOrTupleOfScalars: TypeAlias = extended_typing.MaybeNestedInTuple[core_defs.Scalar]
CompiledProgramsKey: TypeAlias = tuple[
    tuple[ScalarOrTupleOfScalars, ...], common.OffsetProviderType
]


def _hash_compiled_program_unsafe(cp_key: CompiledProgramsKey) -> int:
    values, offset_provider = cp_key
    assert common.is_offset_provider_type(offset_provider)
    return hash((values, id(offset_provider)))


def _validate_types(
    program_name: str,
    static_args: dict[str, ScalarOrTupleOfScalars],
    program_type: ts_ffront.ProgramType,
) -> None:
    unknown_args = list(
        set(static_args.keys()) - set(program_type.definition.pos_or_kw_args.keys())
    )
    if unknown_args:
        raise errors.DSLTypeError(
            message=f"Invalid static arguments provided for '{program_name}' with type '{program_type}', the following are not parameters of the program:\n"
            + ("\n".join([f"  - '{arg}'" for arg in unknown_args])),
            location=None,
        )

    param_types = program_type.definition.pos_or_kw_args

    if type_errors := [
        f"'{name}' with type '{param_types[name]}' cannot be static."
        for name in static_args
        if not type_info.is_type_or_tuple_of_type(param_types[name], ts.ScalarType)
    ]:
        raise errors.DSLTypeError(
            message=f"Invalid static arguments provided for '{program_name}' with type '{program_type}' (only scalars or (nested) tuples of scalars can be static):\n"
            + ("\n".join([f"  - {error}" for error in type_errors])),
            location=None,
        )

    static_arg_types = {name: tt.from_value(value) for name, value in static_args.items()}
    types_from_values = [
        static_arg_types[name]  # use type of the provided value for the static arguments
        if name in static_arg_types
        else type_  # else use the type from progam_type which will never mismatch
        for name, type_ in program_type.definition.pos_or_kw_args.items()
    ]
    assert not program_type.definition.pos_only_args
    assert not program_type.definition.kw_only_args

    if mismatch_errors := list(
        type_info.function_signature_incompatibilities(
            program_type, args=types_from_values, kwargs={}
        )
    ):
        raise errors.DSLTypeError(
            message=f"Invalid static argument types when trying to compile '{program_name}' with type '{program_type}':\n"
            + ("\n".join([f"  - {error}" for error in mismatch_errors])),
            location=None,
        )


def _sanitize_static_args(
    program_name: str,
    static_args: dict[str, ScalarOrTupleOfScalars],
    program_type: ts_ffront.ProgramType,
) -> dict[str, ScalarOrTupleOfScalars]:
    """
    Sanitize static arguments to be used in the program compilation.

    This function will convert all values to their corresponding type
    and check that the types are compatible with the program type.
    """
    _validate_types(program_name, static_args, program_type)

    return {
        name: tt.unsafe_cast_to(value, program_type.definition.pos_or_kw_args[name])  # type: ignore[arg-type] # checked in _validate_types
        for name, value in static_args.items()
    }


@dataclasses.dataclass
class CompiledProgramsPool:
    """
    A pool of compiled programs for a given program and backend.

    If 'static_params' is set (or static arguments are passed to 'compile'),
    the pool will create a program for each argument that is marked static
    and each 'OffsetProviderType'.

    If `enable_jit` is True in the call to the pool, it will compile a program
    with static arguments corresponding to the 'static_params', otherwise it
    will error. In the latter case, the pool needs to be filled with call(s)
    to 'compile' before it can be used.
    """

    backend: gtx_backend.Backend
    definition_stage: ffront_stages.ProgramDefinition
    program_type: ts_ffront.ProgramType
    static_params: Sequence[str] | None = None  # not ordered

    _compiled_programs: eve_utils.CustomMapping = dataclasses.field(
        default_factory=lambda: eve_utils.CustomMapping(_hash_compiled_program_unsafe),
        init=False,
    )

    _offset_provider_type_cache: eve_utils.CustomMapping = dataclasses.field(
        default_factory=lambda: eve_utils.CustomMapping(common.hash_offset_provider_unsafe),
        init=False,
    )  # cache the offset provider type in order to avoid recomputing it at each program call

    def __postinit__(self) -> None:
        # TODO(havogt): We currently don't support pos_only or kw_only args at the program level.
        # This check makes sure we don't miss updating this code if we add support for them in the future.
        assert not self.program_type.definition.kw_only_args
        assert not self.program_type.definition.pos_only_args

    def __call__(
        self, *args: Any, offset_provider: common.OffsetProvider, enable_jit: bool, **kwargs: Any
    ) -> None:
        """
        Calls a program with the given arguments and offset provider.

        If the program is not in cache, it will jit compile with static arguments
        (defined by 'static_params') in case `enable_jit` is True. Otherwise,
        it is an error.
        """
        args, kwargs = type_info.canonicalize_arguments(self.program_type, args, kwargs)
        static_args_values = tuple(args[i] for i in self._static_arg_indices)
        key = (static_args_values, self._offset_provider_to_type_unsafe(offset_provider))
        try:
            self._compiled_programs[key](*args, **kwargs, offset_provider=offset_provider)
        except TypeError:  # 'Future' object is not callable
            # ... otherwise we resolve the future and call again
            program = self._resolve_future(key)
            program(*args, **kwargs, offset_provider=offset_provider)
        except KeyError as e:
            if enable_jit:
                assert self.static_params is not None
                static_args = {
                    name: value
                    for name, value in zip(self.static_params, static_args_values, strict=True)
                }
                self._compile_variant(static_args=static_args, offset_provider=offset_provider)
                return self(
                    *args, offset_provider=offset_provider, enable_jit=False, **kwargs
                )  # passing `enable_jit=False` because a cache miss should be a hard-error in this call`
            raise RuntimeError("No program compiled for this set of static arguments.") from e

    @functools.cached_property
    def _static_arg_indices(self) -> tuple[int, ...]:
        if self.static_params is None:
            # this could also be done in `__call__` but would be an extra if in the fast path
            self.static_params = ()

        all_params = list(self.program_type.definition.pos_or_kw_args.keys())
        return tuple(all_params.index(p) for p in self.static_params)

    def _compile_variant(
        self,
        static_args: dict[str, ScalarOrTupleOfScalars],
        offset_provider: common.OffsetProviderType | common.OffsetProvider,
    ) -> None:
        assert self.static_params is not None
        static_args = _sanitize_static_args(
            self.definition_stage.definition.__name__, static_args, self.program_type
        )

        args = tuple(
            arguments.StaticArg(value=static_args[name], type_=type_)
            if name in self.static_params
            else type_
            for name, type_ in self.program_type.definition.pos_or_kw_args.items()
        )

        key = (
            tuple(static_args[p] for p in self.static_params),
            self._offset_provider_to_type_unsafe(offset_provider),
        )
        if key in self._compiled_programs:
            raise ValueError(f"Program with key {key} already exists.")

        compile_time_args = arguments.CompileTimeArgs(
            offset_provider=offset_provider,  # type:ignore[arg-type] # TODO(havogt): resolve OffsetProviderType vs OffsetProvider
            column_axis=None,  # TODO(havogt): column_axis seems to a unused, even for programs with scans
            args=args,
            kwargs={},
        )
        self._compiled_programs[key] = _async_compilation_pool.submit(
            self.backend.compile, self.definition_stage, compile_time_args=compile_time_args
        )

    def _offset_provider_to_type_unsafe(
        self,
        offset_provider: common.OffsetProvider | common.OffsetProviderType,
    ) -> common.OffsetProviderType:
        try:
            op_type = self._offset_provider_type_cache[offset_provider]
        except KeyError:
            op_type = (
                offset_provider
                if common.is_offset_provider_type(offset_provider)
                else common.offset_provider_to_type(offset_provider)
            )
            self._offset_provider_type_cache[offset_provider] = op_type
        return op_type

    def compile(
        self,
        offset_providers: list[common.OffsetProvider | common.OffsetProviderType],
        **static_args: list[ScalarOrTupleOfScalars],
    ) -> None:
        """
        Compiles the program for all combinations of static arguments and the given 'OffsetProviderType'.

        Note: In case you want to compile for specific combinations of static arguments (instead
        of the combinatoral), you can call compile multiples times.

        Examples:
            pool.compile(static_arg0=[0,1], static_arg1=[2,3], ...)
                will compile for (0,2), (0,3), (1,2), (1,3)
            pool.compile(static_arg0=[0], static_arg1=[2]).compile(static_arg=[1], static_arg1=[3])
                will compile for (0,2), (1,3)
        """
        if self.static_params is None:
            self.static_params = tuple(static_args.keys())
        elif set(self.static_params) != set(static_args.keys()):
            raise ValueError(
                f"Static arguments must be the same for all compiled programs. Got {list(static_args.keys())}, expected {self.static_params}."
            )

        for offset_provider in offset_providers:  # not included in product for better type checking
            for static_values in itertools.product(*static_args.values()):
                self._compile_variant(
                    dict(zip(static_args.keys(), static_values, strict=True)),
                    offset_provider=offset_provider,
                )

    def _resolve_future(self, key: CompiledProgramsKey) -> stages.CompiledProgram:
        program = self._compiled_programs[key]
        assert isinstance(program, concurrent.futures.Future)
        result = program.result()
        self._compiled_programs[key] = result
        return result



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses

import gt4py.next.type_system.type_specifications as ts
from gt4py.eve import codegen
from gt4py.next.otf import languages


def format_source(settings: languages.LanguageSettings, source: str) -> str:
    return codegen.format_source(settings.formatter_key, source, style=settings.formatter_style)


@dataclasses.dataclass(frozen=True)
class Parameter:
    name: str
    type_: ts.TypeSpec


@dataclasses.dataclass(frozen=True)
class Function:
    name: str
    parameters: tuple[Parameter, ...]
    returns: bool = False


@dataclasses.dataclass(frozen=True)
class LibraryDependency:
    name: str
    version: str



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import factory

import gt4py.next.allocators as next_allocators
from gt4py._core import definitions as core_defs
from gt4py.next import backend, common
from gt4py.next.otf import stages, workflow
from gt4py.next.program_processors.runners.dace.workflow.factory import DaCeWorkflowFactory


class DaCeBackendFactory(factory.Factory):
    """
    Workflow factory for the GTIR-DaCe backend.

    Several parameters are inherithed from `backend.Backend`, see below the specific ones.

    Args:
        auto_optimize: Enables the SDFG transformation pipeline.
        make_persistent: Enables optimization in SDFG lowering and bindings generation
            assuming that the layout of temporary and global fields does not change
            across multiple SDFG calls.
    """

    class Meta:
        model = backend.Backend

    class Params:
        name_device = "cpu"
        name_cached = ""
        name_postfix = ""
        gpu = factory.Trait(
            allocator=next_allocators.StandardGPUFieldBufferAllocator(),
            device_type=core_defs.CUPY_DEVICE_TYPE or core_defs.DeviceType.CUDA,
            name_device="gpu",
        )
        cached = factory.Trait(
            executor=factory.LazyAttribute(
                lambda o: workflow.CachedStep(o.otf_workflow, hash_function=o.hash_function)
            ),
            name_cached="_cached",
        )
        device_type = core_defs.DeviceType.CPU
        hash_function = stages.compilation_hash
        otf_workflow = factory.SubFactory(
            DaCeWorkflowFactory,
            device_type=factory.SelfAttribute("..device_type"),
            auto_optimize=factory.SelfAttribute("..auto_optimize"),
            make_persistent=factory.SelfAttribute("..make_persistent"),
        )
        auto_optimize = factory.Trait(name_postfix="_opt")
        make_persistent = False

    name = factory.LazyAttribute(
        lambda o: f"run_dace_{o.name_device}{o.name_cached}{o.name_postfix}"
    )

    executor = factory.LazyAttribute(lambda o: o.otf_workflow)
    allocator = next_allocators.StandardCPUFieldBufferAllocator()
    transforms = backend.DEFAULT_TRANSFORMS


def make_dace_backend(
    auto_optimize: bool,
    cached: bool,
    gpu: bool,
    async_sdfg_call: bool,
    make_persistent: bool,
    use_memory_pool: bool,
    blocking_dim: common.Dimension | None,
    blocking_size: int = 10,
    use_zero_origin: bool = False,
) -> backend.Backend:
    """Helper function to create a dace cached backend with custom config for SDFG
    lowering and auto-optimize.

    Note that `async_sdfg_call=True` on GPU device relies on the dace configuration
    with 1 cuda stream, so that all kernels and memory operations are executed
    sequentially on the default stream.

    Args:
        auto_optimize: Enable SDFG auto-optimize pipeline.
        cached: Cache the lowered SDFG as a JSON file and the compiled programs.
        gpu: Enable GPU transformations and code generation.
        async_sdfg_call: Enable asynchronous SDFG execution, only applicable
            when `gpu=True` because it relies on the gpu kernel queue.
        make_persistent: Allocate persistent arrays with constant layout.
        use_memory_pool: Enable memory pool for allocation of temporary fields.
        blocking_dim: When not 'None', apply 'LoopBlocking' SDFG transformation
            on this dimension.
        blocking_size: Block size to use in 'LoopBlocking' SDFG transformation,
            when enabled.
        use_zero_origin: Can be set to `True` when all fields passed as program
            arguments have zero-based origin. This setting will skip generation
            of range start-symbols `_range_0` since they can be assumed to be zero.

    Returns:
        A custom dace backend object.
    """
    return DaCeBackendFactory(
        gpu=gpu,
        auto_optimize=auto_optimize,
        cached=cached,
        otf_workflow__cached_translation=cached,
        otf_workflow__bare_translation__blocking_dim=blocking_dim,
        otf_workflow__bare_translation__blocking_size=blocking_size,
        otf_workflow__bare_translation__async_sdfg_call=(async_sdfg_call if gpu else False),
        otf_workflow__bare_translation__disable_field_origin_on_program_arguments=use_zero_origin,
        otf_workflow__bare_translation__make_persistent=make_persistent,
        otf_workflow__bare_translation__use_memory_pool=use_memory_pool,
        otf_workflow__bindings__make_persistent=make_persistent,
    )


run_dace_cpu = make_dace_backend(
    auto_optimize=True,
    cached=False,
    gpu=False,
    blocking_dim=None,
    async_sdfg_call=False,
    make_persistent=False,
    use_memory_pool=False,
)
run_dace_cpu_noopt = make_dace_backend(
    auto_optimize=False,
    cached=False,
    gpu=False,
    blocking_dim=None,
    async_sdfg_call=False,
    make_persistent=False,
    use_memory_pool=False,
)
run_dace_cpu_cached = make_dace_backend(
    auto_optimize=True,
    cached=True,
    gpu=False,
    blocking_dim=None,
    async_sdfg_call=False,
    make_persistent=False,
    use_memory_pool=False,
)

run_dace_gpu = make_dace_backend(
    auto_optimize=True,
    cached=False,
    gpu=True,
    blocking_dim=None,
    async_sdfg_call=True,
    make_persistent=False,
    use_memory_pool=True,
)
run_dace_gpu_noopt = make_dace_backend(
    auto_optimize=False,
    cached=False,
    gpu=True,
    blocking_dim=None,
    async_sdfg_call=True,
    make_persistent=False,
    use_memory_pool=True,
)
run_dace_gpu_cached = make_dace_backend(
    auto_optimize=True,
    cached=True,
    gpu=True,
    blocking_dim=None,
    async_sdfg_call=True,
    make_persistent=False,
    use_memory_pool=True,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.iterator import ir
from gt4py.next.iterator.transforms.inline_lifts import _is_lift


def is_eligible_for_inlining(node: ir.FunCall, is_scan_pass_context: bool) -> bool:
    """
    Predicate for the InlineLifts transformation.

    Takes a lifted stencil call of the form `↑(f)(args...)` and returns whether the expression
    is eligible for inlining. For example `·↑(f)(args...)` would be transformed into
    `f(args...)` if true.

    The ``is_scan_pass_context`` argument indicates if the given node is within a scan (unnested),
    e.g. `↑(f)(args...)` should not be inlined if it appears in a scan like this:
    `↑(scan(λ(acc, args...) → acc + ·↑(f)(args...)))(...)`

    Follows the simple rules:
    - Do not inline scans (as there is no efficient way to inline them, also required by some
      backends, e.g. gtfn)
    - Do not inline the first lifted function call within a scan (otherwise, all stencils would get
      inlined into the scans, leading to reduced parallelism/scan-only computation)
    """
    assert _is_lift(node)

    assert isinstance(node.fun, ir.FunCall)  # for mypy
    (stencil,) = node.fun.args
    # Don't inline scans, i.e. exclude `↑(scan(...))(...)`
    if isinstance(stencil, ir.FunCall) and stencil.fun == ir.SymRef(id="scan"):
        return False

    # Don't inline the first lifted function call within a scan, e.g. if the node given here
    # is `↑(f)(args...)` and appears in a scan pass `scan(λ(acc, args...) → acc + ·↑(f)(args...))`
    # it should not be inlined.
    return not is_scan_pass_context



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Any, Generic, Optional, Protocol, TypeAlias, TypeVar

from gt4py.eve import utils
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.otf import arguments, languages, toolchain
from gt4py.next.otf.binding import interface


PrgT = TypeVar("PrgT")
ArgT = TypeVar("ArgT")
SrcL = TypeVar("SrcL", bound=languages.LanguageTag)
TgtL = TypeVar("TgtL", bound=languages.LanguageTag)
SettingT = TypeVar("SettingT", bound=languages.LanguageSettings)
SrcL_co = TypeVar("SrcL_co", bound=languages.LanguageTag, covariant=True)
TgtL_co = TypeVar("TgtL_co", bound=languages.LanguageTag, covariant=True)
SettingT_co = TypeVar("SettingT_co", bound=languages.LanguageSettings, covariant=True)


CompilableProgram: TypeAlias = toolchain.CompilableProgram[itir.Program, arguments.CompileTimeArgs]


def compilation_hash(otf_closure: CompilableProgram) -> int:
    """Given closure compute a hash uniquely determining if we need to recompile."""
    offset_provider = otf_closure.args.offset_provider
    return hash(
        (
            otf_closure.data,
            # As the frontend types contain lists they are not hashable. As a workaround we just
            # use content_hash here.
            utils.content_hash(tuple(arg for arg in otf_closure.args.args)),
            common.hash_offset_provider_unsafe(offset_provider) if offset_provider else None,
            otf_closure.args.column_axis,
        )
    )


def fingerprint_compilable_program(inp: CompilableProgram) -> str:
    """
    Generates a unique hash string for a stencil source program representing
    the program, sorted offset_provider, and column_axis.
    """
    program: itir.Program = inp.data
    offset_provider: common.OffsetProvider = inp.args.offset_provider
    column_axis: Optional[common.Dimension] = inp.args.column_axis

    program_hash = utils.content_hash(
        (
            program,
            sorted(offset_provider.items(), key=lambda el: el[0]),
            column_axis,
        )
    )

    return program_hash


@dataclasses.dataclass(frozen=True)
class ProgramSource(Generic[SrcL, SettingT]):
    """
    Standalone source code translated from an IR along with information relevant for OTF compilation.

    Contains additional information required for further OTF steps, such as
    - implementation language and language conventions
    - dependencies on implementation language libraries
    - how to call the program
    """

    entry_point: interface.Function
    source_code: str
    library_deps: tuple[interface.LibraryDependency, ...]
    language: type[SrcL]
    language_settings: SettingT
    implicit_domain: bool

    def __post_init__(self) -> None:
        if not isinstance(self.language_settings, self.language.settings_class):
            raise TypeError(
                f"Wrong language settings type for '{self.language}', must be subclass of '{self.language.settings_class}'."
            )


@dataclasses.dataclass(frozen=True)
class BindingSource(Generic[SrcL, TgtL]):
    """
    Companion source code for translated program source code.

    This is only needed for OTF compilation if the translated program source code is
    not directly callable from python and therefore requires bindings.
    This can also optionally be added to compile bindings for other languages than python
    when using GT4Py as part of the build for a non-python driver project.
    """

    source_code: str
    library_deps: tuple[interface.LibraryDependency, ...]


# TODO(ricoh): reconsider name in view of future backends producing standalone compilable ProgramSource code
@dataclasses.dataclass(frozen=True)
class CompilableSource(Generic[SrcL, SettingT, TgtL]):
    """
    Encapsulate all the source code required for OTF compilation.

    The bindings module is optional if and only if the program_source is directly callable.
    This should only be the case if the source language / framework supports this out of the box.
    If bindings are required, it is recommended to create them in a separate step to ensure reusability.
    """

    program_source: ProgramSource[SrcL, SettingT]
    binding_source: Optional[BindingSource[SrcL, TgtL]]

    @property
    def library_deps(self) -> tuple[interface.LibraryDependency, ...]:
        if not self.binding_source:
            return self.program_source.library_deps
        return _unique_libs(*self.program_source.library_deps, *self.binding_source.library_deps)


class BuildSystemProject(Protocol[SrcL_co, SettingT_co, TgtL_co]):
    """
    Use source code extracted from a ``CompilableSource`` to configure and build a GT4Py program.

    Should only be considered an OTF stage if used as an endpoint, as this only runs commands on source files
    and is not responsible for importing the results into Python.
    """

    def build(self) -> None: ...


class CompiledProgram(Protocol):
    """Executable python representation of a program."""

    def __call__(self, *args: Any, **kwargs: Any) -> None: ...


class ExtendedCompiledProgram(CompiledProgram, Protocol):
    """Executable python representation of a program with extra info."""

    implicit_domain: bool


def _unique_libs(*args: interface.LibraryDependency) -> tuple[interface.LibraryDependency, ...]:
    """
    Filter out multiple occurrences of the same ``interface.LibraryDependency``.

    Examples:
    ---------
    >>> libs_a = (
    ...     interface.LibraryDependency("foo", "1.2.3"),
    ...     interface.LibraryDependency("common", "1.0.0"),
    ... )
    >>> libs_b = (
    ...     interface.LibraryDependency("common", "1.0.0"),
    ...     interface.LibraryDependency("bar", "1.2.3"),
    ... )
    >>> _unique_libs(*libs_a, *libs_b)
    (LibraryDependency(name='foo', version='1.2.3'), LibraryDependency(name='common', version='1.0.0'), LibraryDependency(name='bar', version='1.2.3'))
    """
    unique: list[interface.LibraryDependency] = []
    for lib in args:
        if lib not in unique:
            unique.append(lib)
    return tuple(unique)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional

import dace

from gt4py._core import definitions as core_defs
from gt4py.next import config


def set_dace_config(
    device_type: core_defs.DeviceType,
    cmake_build_type: Optional[config.CMakeBuildType] = None,
) -> None:
    """Set the DaCe configuration as required by GT4Py.

    This function acts on the current configuration and should not be used inside
    a context, such as `dace.config.temporary_config()`, see note below for more.

    Args:
        device_type: Target device type, needed for compiler config.
        cmake_build_type: CMake build type, needed for compiler config.

    Note:
        Because of the reasons described in [DaCe issue#2125](https://github.com/spcl/dace/issues/2125)
        it is not safe to use this function inside a `dace.config.temporary_config()`
        context or use it to set options that are not static for the process.
    """
    # NOTE: As explained in [DaCe issue#2125](https://github.com/spcl/dace/issues/2125)
    #   it is not possible to use this function inside a configuration context,
    #   or use it to set anything that is not specific to the process. As an example
    #   we can configure which GPU backend is used, under the assumption that we will
    #   only use one type of GPU. Furthermore, we have to set always every time, i.e.
    #   we have to set all flags for the GPU backend even if we are using CPU, since
    #   another thread might work with the GPU right now.

    # We rely on dace cache to avoid recompiling the SDFG.
    #   Note that the workflow step with the persistent `FileCache` store
    #   is translating from `CompilableProgram` (ITIR.Program + CompileTimeArgs)
    #   to `ProgramSource`, so this step is storing in cache only the result
    #   of the SDFG transformations, not the compiled program binary.
    dace.Config.set("cache", value="hash")  # use the SDFG hash as cache key
    dace.Config.set("default_build_folder", value=str(config.BUILD_CACHE_DIR / "dace_cache"))
    dace.Config.set("compiler.use_cache", value=True)

    # Prevents the implicit change of Memlets to Maps. Instead they should be handled by
    #  `gt4py.next.program_processors.runners.dace.transfromations.gpu_utils.gt_gpu_transform_non_standard_memlet()`.
    dace.Config.set("compiler.cuda.allow_implicit_memlet_to_map", value=False)

    if cmake_build_type is not None:
        dace.Config.set("compiler.build_type", value=cmake_build_type.value)

    # dace dafault setting use fast-math in both cpu and gpu compilation, don't use it here
    dace.Config.set(
        "compiler.cpu.args",
        value="-std=c++14 -fPIC -O3 -march=native -Wall -Wextra -Wno-unused-parameter -Wno-unused-label",
    )
    dace.Config.set(
        "compiler.cuda.args",
        value="-Xcompiler -O3 -Xcompiler -march=native -Xcompiler -Wno-unused-parameter",
    )
    dace.Config.set(
        "compiler.cuda.hip_args",
        value="-std=c++17 -fPIC -O3 -march=native -Wno-unused-parameter",
    )

    # In some stencils, for example `apply_diffusion_to_w`, the cuda codegen messes
    #  up with the cuda streams, i.e. it allocates N streams but uses N+1. The first
    #  idea was to use just one stream. However, even in that case the generator
    #  generated wrong code. The current approach is to use the default stream, i.e.
    #  setting `max_concurrent_streams` to `-1`. However, the draw back is, that
    #  apparently then all synchronization is disabled, even the one at the very
    #  end of the SDFG call. To correct for that we are using either
    #  `make_sdfg_call_sync()` or `make_sdfg_call_async()`, see there or in
    #  [DaCe issue#2120](https://github.com/spcl/dace/issues/2120) for more.
    dace.Config.set("compiler.cuda.max_concurrent_streams", value=-1)

    # This assumes that a process will only use one type of GPU.
    if device_type == core_defs.DeviceType.ROCM:
        dace.Config.set("compiler.cuda.backend", value="hip")
    elif device_type == core_defs.DeviceType.CUDA:
        dace.Config.set("compiler.cuda.backend", value="cuda")

    # Instrumentation of SDFG timers
    dace.Config.set("instrumentation", "report_each_invocation", value=True)

    # we are not interested in storing the history of SDFG transformations.
    dace.Config.set("store_history", value=False)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Sequence, Tuple


def dimension_flags_to_names(mask: Tuple[bool, bool, bool]) -> str:
    labels = ["i", "j", "k"]
    selection = [i for i, flag in enumerate(mask) if flag]
    return "".join(labels[i] for i in selection)


def interpolate_mask(seq: Sequence[Any], mask: Sequence[bool], default) -> Tuple[Any, ...]:
    """
    Replace True values by those from the seq in the mask, else default.

    Example:
    >>> default = 0
    >>> a = (1, 2)
    >>> mask = (False, True, False, True)
    >>> interpolate_mask(a, mask, 0)
    (0, 1, 0, 2)
    """
    it = iter(seq)
    return tuple(next(it) if m else default for m in mask)


def filter_mask(seq: Sequence[Any], mask: Sequence[bool]) -> Tuple[Any, ...]:
    """
    Return a reduced-size tuple, with indices where mask[i]=False removed.

    Example:
    >>> a = (1, 2, 3)
    >>> mask = (False, True, False)
    >>> filter_mask(a, mask)
    (2,)
    """
    return tuple(s for m, s in zip(mask, seq) if m)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
The list of exception classes used in the library.

Exception classes that represent errors within an IR go here as a subclass of
:class:`DSLError`. Exception classes that represent other errors, like
the builtin ValueError, go here as well, although you should use Python's
builtin error classes if you can. Exception classes that are specific to a
certain submodule and have no use for the entire application may be better off
in that submodule as opposed to being in this file.
"""

from __future__ import annotations

import textwrap
from typing import Any, Optional

from gt4py.eve import SourceLocation

from . import formatting


class GT4PyError(Exception):
    @property
    def message(self) -> str:
        return self.args[0]


class DSLError(GT4PyError):
    location: Optional[SourceLocation]

    def __init__(self, location: Optional[SourceLocation], message: str) -> None:
        self.location = location
        super().__init__(message)

    def with_location(self, location: Optional[SourceLocation]) -> DSLError:
        self.location = location
        return self

    def __str__(self) -> str:
        if self.location:
            loc_str = formatting.format_location(self.location, show_caret=True)
            return f"{self.message}\n{textwrap.indent(loc_str, '  ')}"
        return self.message


class UnsupportedPythonFeatureError(DSLError):
    feature: str

    def __init__(self, location: Optional[SourceLocation], feature: str) -> None:
        super().__init__(location, f"Unsupported Python syntax: '{feature}'.")
        self.feature = feature


class UndefinedSymbolError(DSLError):
    sym_name: str

    def __init__(self, location: Optional[SourceLocation], name: str) -> None:
        super().__init__(location, f"Name '{name}' is not defined.")
        self.sym_name = name


class MissingAttributeError(DSLError):
    attr_name: str

    def __init__(self, location: Optional[SourceLocation], attr_name: str) -> None:
        super().__init__(location, f"Object does not have attribute '{attr_name}'.")
        self.attr_name = attr_name


class MissingArgumentError(DSLError):
    arg_name: str
    is_kwarg: bool

    def __init__(self, location: Optional[SourceLocation], arg_name: str, is_kwarg: bool) -> None:
        super().__init__(
            location, f"Expected {'keyword-' if is_kwarg else ''}argument '{arg_name}'."
        )
        self.attr_name = arg_name
        self.is_kwarg = is_kwarg


class DSLTypeError(DSLError):
    def __init__(self, location: Optional[SourceLocation], message: str) -> None:
        super().__init__(location, message)


class MissingParameterAnnotationError(DSLTypeError):
    param_name: str

    def __init__(self, location: Optional[SourceLocation], param_name: str) -> None:
        super().__init__(location, f"Parameter '{param_name}' is missing type annotations.")
        self.param_name = param_name


class InvalidParameterAnnotationError(DSLTypeError):
    param_name: str
    annotated_type: Any

    def __init__(self, location: Optional[SourceLocation], param_name: str, type_: Any) -> None:
        super().__init__(
            location, f"Parameter '{param_name}' has invalid type annotation '{type_}'."
        )
        self.param_name = param_name
        self.annotated_type = type_


class CompilationError(GT4PyError):
    def __init__(self, compilation_error: str) -> None:
        super().__init__(f"See attached compilation log.\n{compilation_error}")


class EmbeddedExecutionError(GT4PyError):
    def __init__(self, message: str) -> None:
        super().__init__(message)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import ast
import dataclasses


@dataclasses.dataclass(kw_only=True)
class RemoveDocstrings(ast.NodeTransformer):
    """
    Remove function docstrings from an AST.

    Example
    -------
    >>> import ast, inspect
    >>> def example_docstring():
    ...     a = 1
    ...     "This is a docstring"
    ...
    ...     def example_docstring_2():
    ...         a = 2.0
    ...         "This is a new docstring"
    ...         return a
    ...
    ...     a = example_docstring_2()
    ...     return a
    >>> print(ast.unparse(RemoveDocstrings.apply(ast.parse(inspect.getsource(example_docstring)))))
    def example_docstring():
        a = 1
    <BLANKLINE>
        def example_docstring_2():
            a = 2.0
            return a
        a = example_docstring_2()
        return a
    """

    @classmethod
    def apply(cls, node: ast.AST) -> ast.AST:
        return cls().visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.AST:
        node.body = [obj for obj in node.body if not _is_const_str_expr(obj)]
        return self.generic_visit(node)


def _is_const_str_expr(obj: ast.stmt) -> bool:
    return (
        isinstance(obj, ast.Expr)
        and isinstance(obj.value, ast.Constant)
        and isinstance(obj.value.value, str)
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
from typing import Any, Callable, Iterable, Literal, Mapping, Optional

from gt4py.next import common
from gt4py.next.iterator import builtins, ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.transforms import trace_shifts
from gt4py.next.iterator.transforms.constant_folding import ConstantFolding


def _max_domain_sizes_by_location_type(offset_provider: Mapping[str, Any]) -> dict[str, int]:
    """
    Extract horizontal domain sizes from an `offset_provider`.

    Considers the shape of the neighbor table to get the size of each `source_dim` and the maximum
    value inside the neighbor table to get the size of each `codomain`.
    """
    sizes = dict[str, int]()
    for provider in offset_provider.values():
        if common.is_neighbor_connectivity(provider):
            conn_type = provider.__gt_type__()
            sizes[conn_type.source_dim.value] = max(
                sizes.get(conn_type.source_dim.value, 0), provider.ndarray.shape[0]
            )
            sizes[conn_type.codomain.value] = max(
                sizes.get(conn_type.codomain.value, 0),
                provider.ndarray.max() + 1,  # type: ignore[attr-defined] # TODO(havogt): improve typing for NDArrayObject
            )
    return sizes


@dataclasses.dataclass(frozen=True)
class SymbolicRange:
    start: itir.Expr
    stop: itir.Expr

    def __post_init__(self) -> None:
        # TODO(havogt): added this defensive checks as code seems to make this reasonable assumption
        assert self.start is not itir.InfinityLiteral.POSITIVE
        assert self.stop is not itir.InfinityLiteral.NEGATIVE

    def translate(self, distance: int) -> SymbolicRange:
        return SymbolicRange(im.plus(self.start, distance), im.plus(self.stop, distance))


_GRID_TYPE_MAPPING = {
    "unstructured_domain": common.GridType.UNSTRUCTURED,
    "cartesian_domain": common.GridType.CARTESIAN,
}


@dataclasses.dataclass(frozen=True)
class SymbolicDomain:
    grid_type: common.GridType
    ranges: dict[common.Dimension, SymbolicRange]

    def __hash__(self) -> int:
        return hash((self.grid_type, frozenset(self.ranges.items())))

    @classmethod
    def from_expr(cls, node: itir.Node) -> SymbolicDomain:
        assert cpm.is_call_to(node, ("unstructured_domain", "cartesian_domain"))

        ranges: dict[common.Dimension, SymbolicRange] = {}
        for named_range in node.args:
            assert cpm.is_call_to(named_range, "named_range")
            axis_literal, lower_bound, upper_bound = named_range.args
            assert isinstance(axis_literal, itir.AxisLiteral)

            ranges[common.Dimension(value=axis_literal.value, kind=axis_literal.kind)] = (
                SymbolicRange(lower_bound, upper_bound)
            )
        return cls(_GRID_TYPE_MAPPING[node.fun.id], ranges)

    def as_expr(self) -> itir.FunCall:
        converted_ranges: dict[common.Dimension, tuple[itir.Expr, itir.Expr]] = {
            key: (value.start, value.stop) for key, value in self.ranges.items()
        }
        return im.domain(self.grid_type, converted_ranges)

    def translate(
        self: SymbolicDomain,
        shift: tuple[
            itir.OffsetLiteral
            | Literal[trace_shifts.Sentinel.VALUE, trace_shifts.Sentinel.ALL_NEIGHBORS],
            ...,
        ],
        offset_provider: common.OffsetProvider | common.OffsetProviderType,
        #: A dictionary mapping axes names to their length. See
        #: func:`gt4py.next.iterator.transforms.infer_domain.infer_expr` for more details.
        symbolic_domain_sizes: Optional[dict[str, str]] = None,
    ) -> SymbolicDomain:
        offset_provider_type = common.offset_provider_to_type(offset_provider)

        dims = list(self.ranges.keys())
        new_ranges = {dim: self.ranges[dim] for dim in dims}
        if len(shift) == 0:
            return self
        if len(shift) == 2:
            off, val = shift
            assert isinstance(off, itir.OffsetLiteral) and isinstance(off.value, str)
            connectivity_type = offset_provider_type[off.value]

            if isinstance(connectivity_type, common.Dimension):
                if val is trace_shifts.Sentinel.VALUE:
                    raise NotImplementedError("Dynamic offsets not supported.")
                assert isinstance(val, itir.OffsetLiteral) and isinstance(val.value, int)
                current_dim = connectivity_type
                # cartesian offset
                new_ranges[current_dim] = SymbolicRange.translate(
                    self.ranges[current_dim], val.value
                )
            elif isinstance(connectivity_type, common.NeighborConnectivityType):
                # unstructured shift
                assert (
                    isinstance(val, itir.OffsetLiteral) and isinstance(val.value, int)
                ) or val in [
                    trace_shifts.Sentinel.ALL_NEIGHBORS,
                    trace_shifts.Sentinel.VALUE,
                ]
                horizontal_sizes: dict[str, itir.Expr]
                if symbolic_domain_sizes is not None:
                    horizontal_sizes = {
                        k: im.ensure_expr(v) for k, v in symbolic_domain_sizes.items()
                    }
                else:
                    # note: ugly but cheap re-computation, but should disappear
                    assert common.is_offset_provider(offset_provider)
                    horizontal_sizes = {
                        k: im.literal(str(v), builtins.INTEGER_INDEX_BUILTIN)
                        for k, v in _max_domain_sizes_by_location_type(offset_provider).items()
                    }

                old_dim = connectivity_type.source_dim
                new_dim = connectivity_type.codomain

                assert new_dim not in new_ranges or old_dim == new_dim

                new_range = SymbolicRange(
                    im.literal("0", builtins.INTEGER_INDEX_BUILTIN),
                    horizontal_sizes[new_dim.value],
                )
                new_ranges = dict(
                    (dim, range_) if dim != old_dim else (new_dim, new_range)
                    for dim, range_ in new_ranges.items()
                )
            else:
                raise AssertionError()
            return SymbolicDomain(self.grid_type, new_ranges)
        elif len(shift) > 2:
            return self.translate(shift[0:2], offset_provider, symbolic_domain_sizes).translate(
                shift[2:], offset_provider, symbolic_domain_sizes
            )
        else:
            raise AssertionError("Number of shifts must be a multiple of 2.")


def _reduce_ranges(
    *ranges: SymbolicRange,
    start_reduce_op: Callable[[itir.Expr, itir.Expr], itir.Expr],
    stop_reduce_op: Callable[[itir.Expr, itir.Expr], itir.Expr],
) -> SymbolicRange:
    """Uses start_op and stop_op to fold the start and stop of a list of ranges."""
    start = functools.reduce(
        lambda current_expr, el_expr: start_reduce_op(current_expr, el_expr),
        [range_.start for range_ in ranges],
    )
    stop = functools.reduce(
        lambda current_expr, el_expr: stop_reduce_op(current_expr, el_expr),
        [range_.stop for range_ in ranges],
    )
    # constant fold expression to keep the tree small
    start, stop = ConstantFolding.apply(start), ConstantFolding.apply(stop)  # type: ignore[assignment]  # always an itir.Expr
    return SymbolicRange(start, stop)


_range_union = functools.partial(
    _reduce_ranges, start_reduce_op=im.minimum, stop_reduce_op=im.maximum
)
_range_intersection = functools.partial(
    _reduce_ranges, start_reduce_op=im.maximum, stop_reduce_op=im.minimum
)


def _reduce_domains(
    *domains: SymbolicDomain,
    range_reduce_op: Callable[..., SymbolicRange],
) -> SymbolicDomain:
    """
    Applies range_op to the ranges of a list of domains with same dimensions and grid_type.
    """
    assert all(domain.grid_type == domains[0].grid_type for domain in domains)
    assert all(domain.ranges.keys() == domains[0].ranges.keys() for domain in domains)

    dims = domains[0].ranges.keys()
    new_domain_ranges = {dim: range_reduce_op(*(d.ranges[dim] for d in domains)) for dim in dims}

    return SymbolicDomain(domains[0].grid_type, new_domain_ranges)


domain_union = functools.partial(_reduce_domains, range_reduce_op=_range_union)
"""Return the (set) union of a list of domains."""

domain_intersection = functools.partial(_reduce_domains, range_reduce_op=_range_intersection)
"""Return the intersection of a list of domains."""


def domain_complement(domain: SymbolicDomain) -> SymbolicDomain:
    """
    Return the (set) complement of a half-infinite domain.

    Note: after canonicalization of concat_where, the domain is always half-infinite,
    i.e. it has ranges of the form `]-inf, a[` or `[a, inf[`.
    """
    dims_dict = {}
    for dim in domain.ranges.keys():
        lb, ub = domain.ranges[dim].start, domain.ranges[dim].stop
        assert (lb == itir.InfinityLiteral.NEGATIVE) != (ub == itir.InfinityLiteral.POSITIVE)
        # `]-inf, a[` -> `[a, inf[`
        if lb == itir.InfinityLiteral.NEGATIVE:
            dims_dict[dim] = SymbolicRange(start=ub, stop=itir.InfinityLiteral.POSITIVE)
        # `[a, inf]` -> `]-inf, a]`
        else:  # ub == itir.InfinityLiteral.POSITIVE:
            dims_dict[dim] = SymbolicRange(start=itir.InfinityLiteral.NEGATIVE, stop=lb)
    return SymbolicDomain(domain.grid_type, dims_dict)


def promote_domain(
    domain: SymbolicDomain, target_dims: Iterable[common.Dimension]
) -> SymbolicDomain:
    """Return a domain that is extended with the dimensions of target_dims."""
    assert set(domain.ranges.keys()).issubset(target_dims)
    dims_dict = {
        dim: domain.ranges[dim]
        if dim in domain.ranges
        else SymbolicRange(itir.InfinityLiteral.NEGATIVE, itir.InfinityLiteral.POSITIVE)
        for dim in target_dims
    }
    return SymbolicDomain(domain.grid_type, dims_dict)


def is_finite(range_or_domain: SymbolicRange | SymbolicDomain) -> bool:
    """
    Return whether a range is unbounded in (at least) one direction.

    The expression is required to be constant folded before for the result to be reliable.
    """
    match range_or_domain:
        case SymbolicRange() as range_:
            infinity_literals = (itir.InfinityLiteral.POSITIVE, itir.InfinityLiteral.NEGATIVE)
            return not (range_.start in infinity_literals or range_.stop in infinity_literals)
        case SymbolicDomain() as domain:
            return all(is_finite(range_) for range_ in domain.ranges.values())
        case _:
            raise ValueError("Expected a 'SymbolicRange' or 'SymbolicDomain'.")
